@@ -20,6 +20,26 @@ extern "C" {
 using namespace std;
 
 
+// return as quoted string, with at least '"' backslash-escaped
+template <typename IN> inline string
+lex_cast_qstring(IN const & in)
+{
+  stringstream ss;
+  string out, out2;
+  if (!(ss << in))
+    throw runtime_error("bad lexical cast");
+  out = ss.str();
+  out2 += '"';
+  for (unsigned i=0; i<out.length(); i++)
+    {
+      if (out[i] == '"') // XXX others?
+	out2 += '\\';
+      out2 += out[i];
+    }
+  out2 += '"';
+  return out2;
+}
+
 
 int
 compile_pass (systemtap_session& s)
@@ -30,10 +50,15 @@ compile_pass (systemtap_session& s)
   int rc = 0;
 
   // Create makefile
+
+  for (unsigned i=0; i<s.macros.size(); i++)
+    o << "CFLAGS += -D " << lex_cast_qstring(s.macros[i]) << endl;
+
   if (s.test_mode)
     {
       string module_dir = string("/lib/modules/")
         + s.kernel_release + "/build";
+
       o << "CFLAGS += -I \"" << module_dir << "/include\"" << endl;
       o << "CFLAGS += -I \"" << s.runtime_path << "/user\"" << endl;
       o << "CFLAGS += -I \"" << s.runtime_path << "\"" << endl;
@@ -41,7 +66,6 @@ compile_pass (systemtap_session& s)
       o << s.module_name << ": " << s.translated_source << endl;
       o << "\t$(CC) $(CFLAGS) -o " << s.module_name
         << " " << s.translated_source << endl;
-      o.close ();
     }
   else
     {
@@ -50,13 +74,14 @@ compile_pass (systemtap_session& s)
       o << "CFLAGS += -I \"" << s.runtime_path << "\"" << endl;
       o << "CFLAGS += -I \"" << s.runtime_path << "/relayfs\"" << endl;
       o << "obj-m := " << s.module_name << ".o" << endl;
-      o.close ();
     }
 
+  o.close ();
+
   // Run make
   if (s.test_mode)
     {
-      string make_cmd = string("/usr/bin/make -C \"") + s.tmpdir + "\"";
+      string make_cmd = string("make -C \"") + s.tmpdir + "\"";
 
       if (! s.verbose)
         make_cmd += " -s >/dev/null 2>&1";
@@ -72,7 +97,7 @@ compile_pass (systemtap_session& s)
     {
       string module_dir = string("/lib/modules/")
         + s.kernel_release + "/build";
-      string make_cmd = string("/usr/bin/make")
+      string make_cmd = string("make")
         + string (" -C \"") + module_dir + string("\"");
       make_cmd += string(" M=\"") + s.tmpdir + string("\" modules");
 
@@ -120,10 +145,11 @@ run_pass (systemtap_session& s)
       sighandler_t oldsig = signal (SIGINT, SIG_IGN);
 
       // for now, just spawn stpd
-      string stpd_cmd = string("/usr/bin/sudo ") 
+      string stpd_cmd = string("sudo ") 
         + string(PKGLIBDIR) + "/stpd "
-	+ "-r " // disable relayfs
-        + (s.verbose ? "" : "-q ");
+        + (s.bulk_mode ? "" : "-r ")
+        + (s.verbose ? "" : "-q ")
+        + (s.output_file.empty() ? "" : "-o " + s.output_file + " ");
 
       if (s.cmd != "")
 	stpd_cmd += "-c \"" + s.cmd + "\" ";
@@ -131,6 +157,9 @@ run_pass (systemtap_session& s)
       if (s.target_pid)
 	stpd_cmd += "-t " + stringify(s.target_pid) + " ";
 
+      if (s.buffer_size)
+	stpd_cmd += "-b " + stringify(s.buffer_size) + " ";
+
       stpd_cmd += s.tmpdir + "/" + s.module_name + ".ko";
 
       if (s.verbose) clog << "Running " << stpd_cmd << endl;
