@@ -24,6 +24,17 @@ extern "C" {
 using namespace std;
 
 
+template <typename OUT, typename IN> inline OUT
+lex_cast(IN const & in)
+{
+  stringstream ss;
+  OUT out;
+  if (!(ss << in && ss >> out))
+    throw runtime_error("bad lexical cast");
+  return out;
+}
+
+
 // ------------------------------------------------------------------------
 
 
@@ -140,10 +151,10 @@ match_key::operator<(match_key const & other) const
 {
   return ((name < other.name)
 	  
-	  || (name == name 
+	  || (name == other.name 
 	      && have_parameter < other.have_parameter)
 	  
-	  || (name == name 
+	  || (name == other.name 
 	      && have_parameter == other.have_parameter 
 	      && parameter_type < other.parameter_type));
 }
@@ -159,6 +170,9 @@ match_node::match_node()
 match_node *
 match_node::bind(match_key const & k) 
 {
+  if (k.name == "*")
+    throw semantic_error("invalid use of wildcard probe point component");
+
   map<match_key, match_node *>::const_iterator i = sub.find(k);
   if (i != sub.end())
     return i->second;
@@ -171,7 +185,7 @@ void
 match_node::bind(derived_probe_builder * e)
 {
   if (end)
-    throw semantic_error("already have a pattern ending");
+    throw semantic_error("duplicate probe point pattern");
   end = e;
 }
 
@@ -193,70 +207,89 @@ match_node::bind_num(string const & k)
   return bind(match_key(k).with_number());
 }
 
-derived_probe_builder * 
-match_node::find_builder(vector<probe_point::component *> const & components,
-			 unsigned pos,
-			 vector< pair<string, literal *> > & parameters)
+
+void
+match_node::find_and_build (systemtap_session& s,
+                            probe* p, probe_point *loc, unsigned pos,
+                            vector<derived_probe *>& results)
 {
-  assert(pos <= components.size());
-  if (pos == components.size())
+  assert (pos <= loc->components.size());
+  if (pos == loc->components.size()) // matched all probe point components so far 
     {
-      // Probe_point ends here. We match iff we have
-      // an "end" entry here. If we don't, it'll be null.
-      return end;
+      derived_probe_builder *b = end; // may be 0 if only nested names are bound
+
+      if (! b)
+        {
+          string alternatives;
+          for (sub_map_iterator_t i = sub.begin(); i != sub.end(); i++)
+            alternatives += string(" ") + i->first.str();
+
+          throw semantic_error (string("probe point truncated at position ") + 
+                                lex_cast<string> (pos) +
+                                " (follow:" + alternatives + ")");
+        }
+
+      map<string, literal *> param_map;
+      for (unsigned i=0; i<pos; i++)
+        param_map[loc->components[i]->functor] = loc->components[i]->arg;
+      // maybe 0
+
+      b->build (s, p, loc, param_map, results);
     }
-  else
+  else if (loc->components[pos]->functor == "*") // wildcard?
     {
-      // Probe_point contains a component here. We match iff there's
-      // an entry in the sub table, and its value matches the rest
-      // of the probe_point.
-      match_key k(*components[pos]);
-      if (0)
-	clog << "searching for component " << k.str() << endl;
-      map<match_key, match_node *>::const_iterator i = sub.find(k);
-      if (i == sub.end())
-	{
-	  if (0)
-	    clog << "no match found" << endl;
-	  return NULL;
-	}
-      else
-	{
-	  if (0)
-	    clog << "matched " << k.str() << endl;
-	  derived_probe_builder * builder = NULL;
-	  if (k.have_parameter)
-	    {
-	      assert(components[pos]->arg);
-	      parameters.push_back(make_pair(components[pos]->functor, 
-					     components[pos]->arg));
-	    }
-	  else
-	    {
-	      // store a "null parameter" for any component we run into, anyways
-	      literal_string *empty = NULL;
-	      parameters.push_back(make_pair(components[pos]->functor, empty));
-	    }
-	  builder = i->second->find_builder(components, pos+1, parameters);
-	  if (k.have_parameter && !builder)
-	    parameters.pop_back();
-	  return builder;
-	}
+      // Recursively call derive_probes for all matches of the current
+      // key position.  To do this, we have to perform almost an
+      // alias-level duplication of the probe body and synthesis of
+      // new probe_points.
+
+      probe * n = new probe();
+      n->tok = p->tok;
+      n->body = deep_copy_visitor::deep_copy(p->body);
+
+      // Construct N probe_point instances: one per wildcard match
+      for (sub_map_iterator_t i = sub.begin(); i != sub.end(); i++)
+        {
+          const match_key& match = i->first;
+          probe_point *loc2 = new probe_point;
+          loc2->tok = loc->tok;
+
+          // deep-copy probe point components, except for this wildcard position
+          for (unsigned k=0; k<loc->components.size(); k++)
+            if (pos != k)
+              loc2->components.push_back(new probe_point::component(loc->components[k]->functor,
+                                                                    loc->components[k]->arg));
+            else
+              // NB: we retain wildcard arg - foo.*(5).bar => foo.a(5).bar, foo.b(5).bar
+              loc2->components.push_back(new probe_point::component(match.name,
+                                                                    loc->components[k]->arg));
+          n->locations.push_back (loc2);
+        }
+
+      derive_probes (s, n, results, false);
+    }
+  else 
+    {
+      match_key match (* loc->components[pos]);
+      sub_map_iterator_t i = sub.find (match);
+      if (i == sub.end()) // no match
+        {
+          string alternatives;
+          for (sub_map_iterator_t i = sub.begin(); i != sub.end(); i++)
+            alternatives += string(" ") + i->first.str();
+          
+          throw semantic_error (string("probe point mismatch at position ") + 
+                                lex_cast<string> (pos) +
+                                " (alternatives:" + alternatives + ")");
+        }
+
+      match_node* subnode = i->second;
+      // recurse
+      subnode->find_and_build (s, p, loc, pos+1, results);
     }
 }
 
 
-static void
-param_vec_to_map(vector< pair<string, literal *> > const & param_vec, 
-		   map<string, literal *> & param_map)
-{
-  for (vector< pair<string, literal *> >::const_iterator i = param_vec.begin();
-       i != param_vec.end(); ++i)
-    {
-      param_map[i->first] = i->second;
-    }
-}
-
 // ------------------------------------------------------------------------
 // Alias probes
 // ------------------------------------------------------------------------
@@ -275,7 +308,6 @@ alias_expansion_builder
 		     probe * use, 
 		     probe_point * location,
 		     std::map<std::string, literal *> const & parameters,
-		     vector<probe *> & results_to_expand_further,
 		     vector<derived_probe *> & finished_results)
   {
     // We're going to build a new probe and wrap it up in an
@@ -311,8 +343,8 @@ alias_expansion_builder
 	statement *s = deep_copy_visitor::deep_copy(use->body->statements[i]);
 	n->body->statements.push_back(s);
       }
-  
-    results_to_expand_further.push_back(n);
+
+    derive_probes (sess, n, finished_results, false);
   }
 };
 
@@ -337,29 +369,32 @@ systemtap_session::register_library_aliases()
       for (unsigned a = 0; a < file->aliases.size(); ++a)
 	{
 	  probe_alias * alias = file->aliases[a];
-	  for (unsigned n = 0; n < alias->alias_names.size(); ++n)
-	    {
-	      probe_point * name = alias->alias_names[n];
-	      if (verbose)
-		{
-		  clog << "registering probe alias ";
-		  for (unsigned c = 0; c < name->components.size(); ++c)
-		    clog << (c > 0 ? "." : "") << name->components[c]->functor;
-		  clog << endl;
-		}
-	      match_node * n = pattern_root;
-	      for (unsigned c = 0; c < name->components.size(); ++c)
-		{
-		  probe_point::component * comp = name->components[c];
-		  // XXX: alias parameters
-		  if (comp->arg)
-		    throw semantic_error("alias component " 
-					 + comp->functor 
-					 + " contains illegal parameter");
-		  n = n->bind(comp->functor);
-		}
-	      n->bind(new alias_expansion_builder(alias));
-	    }
+          try 
+            {
+              for (unsigned n = 0; n < alias->alias_names.size(); ++n)
+                {
+                  probe_point * name = alias->alias_names[n];
+                  match_node * n = pattern_root;
+                  for (unsigned c = 0; c < name->components.size(); ++c)
+                    {
+                      probe_point::component * comp = name->components[c];
+                      // XXX: alias parameters
+                      if (comp->arg)
+                        throw semantic_error("alias component " 
+                                             + comp->functor 
+                                             + " contains illegal parameter");
+                      n = n->bind(comp->functor);
+                    }
+                  n->bind(new alias_expansion_builder(alias));
+                }
+            }
+          catch (const semantic_error& e)
+            {
+              print_error (e);
+              cerr << "         while: registering probe alias ";
+              alias->printsig(cerr);
+              cerr << endl;
+            }
 	}
     }
 }
@@ -385,101 +420,153 @@ recursion_guard
 
 // The match-and-expand loop.
 void
-symresolution_info::derive_probes (match_node * root, 
-				   probe *p, vector<derived_probe*>& dps)
+derive_probes (systemtap_session& s,
+               probe *p, vector<derived_probe*>& dps,
+               bool exc_outermost)
 {
-  static unsigned depth=0;
-  recursion_guard guard(depth);
-
   for (unsigned i = 0; i < p->locations.size(); ++i)
     {
       probe_point *loc = p->locations[i];
-      vector< pair<string, literal *> > param_vec;
-      map<string, literal *> param_map;
-      vector<probe *> re_expand;
-
+      
       try
         {
-          derived_probe_builder * builder = 
-            root->find_builder(loc->components, 0, param_vec);
-          
-          if (!builder)
-            throw semantic_error ("no match for probe point family");
-          
-          param_vec_to_map(param_vec, param_map);
-          
           unsigned num_atbegin = dps.size();
-          builder->build(session, p, loc, param_map, re_expand, dps);
-          
-          // Recursively expand any further-expanding results
-          // XXX: ... or the build routine could get a reference to
-          // this object and let it call derive_probes() internally
-          if (!re_expand.empty())
-            {
-              for (unsigned j = 0; j < re_expand.size(); ++j)
-                derive_probes(root, re_expand[j], dps);
-            }
-          
+          s.pattern_root->find_and_build (s, p, loc, 0, dps);
           unsigned num_atend = dps.size();
+          
           if (num_atbegin == num_atend) // nothing new derived!
             throw semantic_error ("no match for probe point");
         }
       catch (const semantic_error& e)
         {
-          session.print_error (e);
+          // XXX: prefer not to print_error at every nest/unroll level
+          s.print_error (e);
           cerr << "         while: resolving probe point " << *loc << endl;
+          //          if (! exc_outermost)
+          //            throw;
         }
     }
 }
 
+
+
 // ------------------------------------------------------------------------
 //
-// Map usage checks
+// Indexable usage checks
 //
 
+struct symbol_fetcher
+  : public throwing_visitor
+{
+  symbol *&sym;
 
-struct mutated_map_collector
+  symbol_fetcher (symbol *&sym)
+    : sym(sym) 
+  {}
+
+  void visit_symbol (symbol* e)
+  {
+    sym = e;
+  }
+
+  void visit_arrayindex (arrayindex* e)
+  {
+    e->base->visit_indexable (this);
+  }
+
+  void throwone (const token* t)
+  {
+    throw semantic_error ("Expecting symbol or array index expression", t);
+  }
+};
+
+symbol *
+get_symbol_within_expression (expression *e)
+{
+  symbol *sym = NULL;
+  symbol_fetcher fetcher(sym);
+  e->visit (&fetcher);
+  if (!sym)
+    throw semantic_error("Unable to find symbol in expression", e->tok);
+  return sym;
+}
+
+static symbol *
+get_symbol_within_indexable (indexable *ix)
+{
+  symbol *array = NULL;
+  hist_op *hist = NULL;
+  classify_indexable(ix, array, hist);
+  if (array)
+    return array;
+  else
+    return get_symbol_within_expression (hist->stat);
+}
+
+struct mutated_var_collector
   : public traversing_visitor
 {
-  set<vardecl *> * mutated_maps;
+  set<vardecl *> * mutated_vars;
 
-  mutated_map_collector(set<vardecl *> * mm) 
-    : mutated_maps (mm)
+  mutated_var_collector (set<vardecl *> * mm) 
+    : mutated_vars (mm)
   {}
 
+  void visit_assignment(assignment* e)
+  {
+    if (e->type == pe_stats && e->op == "<<<")
+      {
+	vardecl *vd = get_symbol_within_expression (e->left)->referent;
+	if (vd)
+	  mutated_vars->insert (vd);
+      }
+    e->left->visit (this);
+    e->right->visit (this);
+  }
+
   void visit_arrayindex (arrayindex *e)
   {
-    if (is_active_lvalue(e))
-      mutated_maps->insert(e->referent);
+    if (is_active_lvalue (e))
+      {
+	symbol *sym;
+	if (e->base->is_symbol (sym))
+	  mutated_vars->insert (sym->referent);
+	else
+	  throw semantic_error("Assignment to read-only histogram bucket", e->tok);
+      }
   }
 };
 
 
-struct no_map_mutation_during_iteration_check
+struct no_var_mutation_during_iteration_check
   : public traversing_visitor
 {
   systemtap_session & session;
-  map<functiondecl *,set<vardecl *> *> & function_mutates_maps;
-  vector<vardecl *> maps_being_iterated;
+  map<functiondecl *,set<vardecl *> *> & function_mutates_vars;
+  vector<vardecl *> vars_being_iterated;
   
-  no_map_mutation_during_iteration_check 
+  no_var_mutation_during_iteration_check 
   (systemtap_session & sess,
-   map<functiondecl *,set<vardecl *> *> & fmm)
-    : session(sess), function_mutates_maps (fmm)
+   map<functiondecl *,set<vardecl *> *> & fmv)
+    : session(sess), function_mutates_vars (fmv)
   {}
 
   void visit_arrayindex (arrayindex *e)
   {
     if (is_active_lvalue(e))
       {
-	for (unsigned i = 0; i < maps_being_iterated.size(); ++i)
+	vardecl *vd = get_symbol_within_indexable (e->base)->referent;
+	if (vd)
 	  {
-	    vardecl *m = maps_being_iterated[i];
-	    if (m == e->referent)
+	    for (unsigned i = 0; i < vars_being_iterated.size(); ++i)
 	      {
-		string err = ("map '" + m->name +
-			      "' modified during 'foreach' iteration");
-		session.print_error (semantic_error (err, e->tok));
+		vardecl *v = vars_being_iterated[i];
+		if (v == vd)
+		  {
+		    string err = ("variable '" + v->name +
+				  "' modified during 'foreach' iteration");
+		    session.print_error (semantic_error (err, e->tok));
+		  }
 	      }
 	  }
       }
@@ -488,16 +575,16 @@ struct no_map_mutation_during_iteration_check
   void visit_functioncall (functioncall* e)
   {
     map<functiondecl *,set<vardecl *> *>::const_iterator i 
-      = function_mutates_maps.find (e->referent);
+      = function_mutates_vars.find (e->referent);
 
-    if (i != function_mutates_maps.end())
+    if (i != function_mutates_vars.end())
       {
-	for (unsigned j = 0; j < maps_being_iterated.size(); ++j)
+	for (unsigned j = 0; j < vars_being_iterated.size(); ++j)
 	  {
-	    vardecl *m = maps_being_iterated[j];
+	    vardecl *m = vars_being_iterated[j];
 	    if (i->second->find (m) != i->second->end())
 	      {
-		string err = ("function call modifies map '" + m->name +
+		string err = ("function call modifies var '" + m->name +
 			      "' during 'foreach' iteration");
 		session.print_error (semantic_error (err, e->tok));
 	      }
@@ -510,21 +597,27 @@ struct no_map_mutation_during_iteration_check
 
   void visit_foreach_loop(foreach_loop* s)
   {
-    maps_being_iterated.push_back (s->base_referent);
+    vardecl *vd = get_symbol_within_indexable (s->base)->referent;
+
+    if (vd)
+      vars_being_iterated.push_back (vd);
+    
     for (unsigned i=0; i<s->indexes.size(); i++)
       s->indexes[i]->visit (this);
     s->block->visit (this);
-    maps_being_iterated.pop_back();
+
+    if (vd)
+      vars_being_iterated.pop_back();
   }
 };
 
 
 static int
-semantic_pass_maps (systemtap_session & sess)
+semantic_pass_vars (systemtap_session & sess)
 {
   
-  map<functiondecl *, set<vardecl *> *> fmm;
-  no_map_mutation_during_iteration_check chk(sess, fmm);
+  map<functiondecl *, set<vardecl *> *> fmv;
+  no_var_mutation_during_iteration_check chk(sess, fmv);
   
   for (unsigned i = 0; i < sess.functions.size(); ++i)
     {
@@ -532,9 +625,9 @@ semantic_pass_maps (systemtap_session & sess)
       if (fn->body)
 	{
 	  set<vardecl *> * m = new set<vardecl *>();
-	  mutated_map_collector mc (m);
+	  mutated_var_collector mc (m);
 	  fn->body->visit (&mc);
-	  fmm[fn] = m;
+	  fmv[fn] = m;
 	}
     }
 
@@ -555,10 +648,113 @@ semantic_pass_maps (systemtap_session & sess)
 
 // ------------------------------------------------------------------------
 
+struct stat_decl_collector
+  : public traversing_visitor
+{
+  systemtap_session & session;
+  
+  stat_decl_collector(systemtap_session & sess)
+    : session(sess)
+  {}
+
+  void visit_stat_op (stat_op* e)
+  {
+    symbol *sym = get_symbol_within_expression (e->stat);
+    if (session.stat_decls.find(sym->name) == session.stat_decls.end())
+      session.stat_decls[sym->name] = statistic_decl();
+  }
+
+  void visit_assignment (assignment* e)
+  {
+    if (e->op == "<<<")
+      {
+	symbol *sym = get_symbol_within_expression (e->left);
+	if (session.stat_decls.find(sym->name) == session.stat_decls.end())
+	  session.stat_decls[sym->name] = statistic_decl();
+      }
+    else
+      traversing_visitor::visit_assignment(e);
+  }
+
+  void visit_hist_op (hist_op* e)
+  {
+    symbol *sym = get_symbol_within_expression (e->stat);
+    statistic_decl new_stat;
+
+    if (e->htype == hist_linear)
+      {
+	new_stat.type = statistic_decl::linear;
+	assert (e->params.size() == 3);
+	new_stat.linear_low = e->params[0];
+	new_stat.linear_high = e->params[1];
+	new_stat.linear_step = e->params[2];
+      }
+    else
+      {
+	assert (e->htype == hist_log);
+	new_stat.type = statistic_decl::logarithmic;
+	assert (e->params.size() == 1);
+	new_stat.logarithmic_buckets = e->params[0];
+      }
+
+    map<string, statistic_decl>::iterator i = session.stat_decls.find(sym->name);
+    if (i == session.stat_decls.end())
+      session.stat_decls[sym->name] = new_stat;
+    else
+      {
+	statistic_decl & old_stat = i->second;
+	if (!(old_stat == new_stat))
+	  {
+	    if (old_stat.type == statistic_decl::none)
+	      i->second = new_stat;
+	    else
+	      {
+		// FIXME: Support multiple co-declared histogram types
+		semantic_error se("multiple histogram types declared on '" + sym->name + "'", 
+				  e->tok);
+		session.print_error (se);
+	      }
+	  }
+      }    
+  }
+
+};
+
+static int
+semantic_pass_stats (systemtap_session & sess)
+{
+  stat_decl_collector sdc(sess);
+
+  for (unsigned i = 0; i < sess.functions.size(); ++i)    
+    sess.functions[i]->body->visit (&sdc);
+
+  for (unsigned i = 0; i < sess.probes.size(); ++i)    
+    sess.probes[i]->body->visit (&sdc);
+
+  for (unsigned i = 0; i < sess.globals.size(); ++i)
+    {
+      vardecl *v = sess.globals[i];
+      if (v->type == pe_stats)
+	{
+	  
+	  if (sess.stat_decls.find(v->name) == sess.stat_decls.end())
+	    {
+	      semantic_error se("unable to infer statistic parameters for global '" + v->name + "'");
+	      sess.print_error (se);
+	    }
+	}
+    }
+  
+  return sess.num_errors;
+}
+
+// ------------------------------------------------------------------------
+
 
 static int semantic_pass_symbols (systemtap_session&);
 static int semantic_pass_types (systemtap_session&);
-static int semantic_pass_maps (systemtap_session&);
+static int semantic_pass_vars (systemtap_session&);
+static int semantic_pass_stats (systemtap_session&);
 
 
 
@@ -619,7 +815,7 @@ semantic_pass_symbols (systemtap_session& s)
 
           // much magic happens here: probe alias expansion,
           // provider identification
-          sym.derive_probes (s.pattern_root, p, dps);
+          derive_probes (s, p, dps);
 
           for (unsigned j=0; j<dps.size(); j++)
             {
@@ -648,12 +844,23 @@ semantic_pass_symbols (systemtap_session& s)
 int
 semantic_pass (systemtap_session& s)
 {
-  s.register_library_aliases();
-  register_standard_tapsets(s);
+  int rc = 0;
 
-  int rc = semantic_pass_symbols (s);
-  if (rc == 0) rc = semantic_pass_types (s);
-  if (rc == 0) rc = semantic_pass_maps (s);
+  try 
+    {
+      s.register_library_aliases();
+      register_standard_tapsets(s);
+      
+      rc = semantic_pass_symbols (s);
+      if (rc == 0) rc = semantic_pass_types (s);
+      if (rc == 0) rc = semantic_pass_vars (s);
+      if (rc == 0) rc = semantic_pass_stats (s);
+    }
+  catch (const semantic_error& e)
+    {
+      s.print_error (e);
+    }
+  
   return rc;
 }
 
@@ -715,18 +922,31 @@ symresolution_info::visit_foreach_loop (foreach_loop* e)
   for (unsigned i=0; i<e->indexes.size(); i++)
     e->indexes[i]->visit (this);
 
-  if (e->base_referent)
-    return;
+  symbol *array = NULL;  
+  hist_op *hist = NULL;
+  classify_indexable (e->base, array, hist);
 
-  vardecl* d = find_var (e->base, e->indexes.size ());
-  if (d)
-    e->base_referent = d;
-  else
-    throw semantic_error ("unresolved global array " + e->base, e->tok);
+  if (array)
+    {
+      if (!array->referent)
+	{	  
+	  vardecl* d = find_var (array->name, e->indexes.size ());
+	  if (d)
+	    array->referent = d;
+	  else
+	    throw semantic_error ("unresolved global array " + array->name, e->tok);
+	}
+    }
+  else 
+    {
+      assert (hist);
+      hist->visit (this);
+    }
 
   e->block->visit (this);
 }
 
+
 struct 
 delete_statement_symresolution_info:
   public traversing_visitor
@@ -800,27 +1020,39 @@ symresolution_info::visit_arrayindex (arrayindex* e)
   for (unsigned i=0; i<e->indexes.size(); i++)
     e->indexes[i]->visit (this);
 
-  if (e->referent)
-    return;
+  symbol *array = NULL;  
+  hist_op *hist = NULL;
+  classify_indexable(e->base, array, hist);
 
-  vardecl* d = find_var (e->base, e->indexes.size ());
-  if (d)
-    e->referent = d;
+  if (array)
+    {
+      if (array->referent)
+	return;
+
+      vardecl* d = find_var (array->name, e->indexes.size ());
+      if (d)
+	array->referent = d;
+      else
+	{
+	  // new local
+	  vardecl* v = new vardecl;
+	  v->set_arity(e->indexes.size());
+	  v->name = array->name;
+	  v->tok = array->tok;
+	  if (current_function)
+	    current_function->locals.push_back (v);
+	  else if (current_probe)
+	    current_probe->locals.push_back (v);
+	  else
+	    // must not happen
+	    throw semantic_error ("no current probe/function", e->tok);
+	  array->referent = v;
+	}      
+    }
   else
     {
-      // new local
-      vardecl* v = new vardecl;
-      v->set_arity(e->indexes.size());
-      v->name = e->base;
-      v->tok = e->tok;
-      if (current_function)
-        current_function->locals.push_back (v);
-      else if (current_probe)
-        current_probe->locals.push_back (v);
-      else
-        // must not happen
-        throw semantic_error ("no current probe/function", e->tok);
-      e->referent = v;
+      assert (hist);
+      hist->visit (this);
     }
 }
 
@@ -1103,12 +1335,20 @@ typeresolution_info::visit_assignment (assignment *e)
       e->left->visit (this);
       t = pe_long;
       e->right->visit (this);
-      if (e->type == pe_unknown)
+      if (e->type == pe_unknown ||
+	  e->type == pe_stats)
         {
           e->type = pe_long;
           resolved (e->tok, e->type);
         }
     }
+
+  else if (e->left->type == pe_stats)
+    invalid (e->left->tok, e->left->type);
+
+  else if (e->right->type == pe_stats)
+    invalid (e->right->tok, e->right->type);
+
   else if (e->op == "+=" || // numeric only
            e->op == "-=" ||
            e->op == "*=" ||
@@ -1172,6 +1412,7 @@ typeresolution_info::visit_assignment (assignment *e)
           e->right->type != pe_unknown &&
           e->left->type != e->right->type)
         mismatch (e->tok, e->left->type, e->right->type);
+
     }
   else
     throw semantic_error ("unsupported assignment operator " + e->op);
@@ -1321,11 +1562,7 @@ void
 typeresolution_info::visit_symbol (symbol* e)
 {
   assert (e->referent != 0);
-
-  if (e->referent->arity > 0)
-    unresolved (e->tok); // symbol resolution should not permit this
-  else
-    resolve_2types (e, e->referent, this, t);
+  resolve_2types (e, e->referent, this, t);
 }
 
 
@@ -1339,9 +1576,32 @@ typeresolution_info::visit_target_symbol (target_symbol* e)
 void
 typeresolution_info::visit_arrayindex (arrayindex* e)
 {
-  assert (e->referent != 0);
 
-  resolve_2types (e, e->referent, this, t);
+  symbol *array = NULL;
+  hist_op *hist = NULL;
+  classify_indexable(e->base, array, hist);
+  
+  // Every hist_op has type [int]:int, that is to say, every hist_op
+  // is a pseudo-one-dimensional integer array type indexed by
+  // integers (bucket numbers).
+
+  if (hist)
+    {
+      if (e->indexes.size() != 1)
+	unresolved (e->tok);
+      t = pe_long;
+      e->indexes[0]->visit (this);
+      if (e->indexes[0]->type != pe_long)
+	unresolved (e->tok);
+      hist->visit (this);
+      return;
+    }
+
+  // Now we are left with "normal" map inference and index checking.
+
+  assert (array);
+  assert (array->referent != 0);
+  resolve_2types (e, array->referent, this, t);
 
   // now resolve the array indexes
 
@@ -1349,12 +1609,12 @@ typeresolution_info::visit_arrayindex (arrayindex* e)
   //   // redesignate referent as array
   //   e->referent->set_arity (e->indexes.size ());
 
-  if (e->indexes.size() != e->referent->index_types.size())
+  if (e->indexes.size() != array->referent->index_types.size())
     unresolved (e->tok); // symbol resolution should prevent this
   else for (unsigned i=0; i<e->indexes.size(); i++)
     {
       expression* ee = e->indexes[i];
-      exp_type& ft = e->referent->index_types [i];
+      exp_type& ft = array->referent->index_types [i];
       t = ft;
       ee->visit (this);
       exp_type at = ee->type;
@@ -1363,7 +1623,7 @@ typeresolution_info::visit_arrayindex (arrayindex* e)
         {
           // propagate to formal type
           ft = at;
-          resolved (e->referent->tok, ft);
+          resolved (array->referent->tok, ft);
           // uses array decl as there is no token for "formal type"
         }
       if (at == pe_stats)
@@ -1487,31 +1747,49 @@ typeresolution_info::visit_foreach_loop (foreach_loop* e)
   //   // redesignate referent as array
   //   e->referent->set_arity (e->indexes.size ());
 
-  if (e->indexes.size() != e->base_referent->index_types.size())
-    unresolved (e->tok); // symbol resolution should prevent this
-  else for (unsigned i=0; i<e->indexes.size(); i++)
-    {
-      expression* ee = e->indexes[i];
-      exp_type& ft = e->base_referent->index_types [i];
-      t = ft;
-      ee->visit (this);
-      exp_type at = ee->type;
+  symbol *array = NULL;
+  hist_op *hist = NULL;
+  classify_indexable(e->base, array, hist);
 
-      if ((at == pe_string || at == pe_long) && ft == pe_unknown)
-        {
-          // propagate to formal type
-          ft = at;
-          resolved (e->base_referent->tok, ft);
-          // uses array decl as there is no token for "formal type"
-        }
-      if (at == pe_stats)
-        invalid (ee->tok, at);
-      if (ft == pe_stats)
-        invalid (ee->tok, ft);
-      if (at != pe_unknown && ft != pe_unknown && ft != at)
-        mismatch (e->tok, at, ft);
-      if (at == pe_unknown)
-        unresolved (ee->tok);
+  if (hist)
+    {      
+      if (e->indexes.size() != 1)
+	unresolved (e->tok);
+      t = pe_long;
+      e->indexes[0]->visit (this);
+      if (e->indexes[0]->type != pe_long)
+	unresolved (e->tok);
+      hist->visit (this);
+    }
+  else
+    {
+      assert (array);  
+      if (e->indexes.size() != array->referent->index_types.size())
+	unresolved (e->tok); // symbol resolution should prevent this
+      else for (unsigned i=0; i<e->indexes.size(); i++)
+	{
+	  expression* ee = e->indexes[i];
+	  exp_type& ft = array->referent->index_types [i];
+	  t = ft;
+	  ee->visit (this);
+	  exp_type at = ee->type;
+	  
+	  if ((at == pe_string || at == pe_long) && ft == pe_unknown)
+	    {
+	      // propagate to formal type
+	      ft = at;
+	      resolved (array->referent->tok, ft);
+	      // uses array decl as there is no token for "formal type"
+	    }
+	  if (at == pe_stats)
+	    invalid (ee->tok, at);
+	  if (ft == pe_stats)
+	    invalid (ee->tok, ft);
+	  if (at != pe_unknown && ft != pe_unknown && ft != at)
+	    mismatch (e->tok, at, ft);
+	  if (at == pe_unknown)
+	    unresolved (ee->tok);
+	}
     }
 
   t = pe_unknown;
@@ -1631,6 +1909,136 @@ typeresolution_info::visit_return_statement (return_statement* e)
     invalid (e->value->tok, e->value->type);
 }
 
+void 
+typeresolution_info::visit_print_format (print_format* e)
+{
+  size_t unresolved_args = 0;
+
+  if (e->print_with_format)
+    {
+      // If there's a format string, we can do both inference *and*
+      // checking.
+
+      // First we extract the subsequence of formatting components
+      // which are conversions (not just literal string components)
+
+      std::vector<print_format::format_component> components;
+      for (size_t i = 0; i < e->components.size(); ++i)
+	{
+	  if (e->components[i].type == print_format::conv_unspecified)
+	    throw semantic_error ("Unspecified conversion in print operator format string",
+				  e->tok);
+	  else if (e->components[i].type == print_format::conv_literal)
+	    continue;
+	  components.push_back(e->components[i]);
+	}
+
+      // Then we check that the number of conversions and the number
+      // of args agree.
+
+      if (components.size() != e->args.size())
+	throw semantic_error ("Wrong number of args to formatted print operator",
+			      e->tok);
+
+      // Then we check that the types of the conversions match the types
+      // of the args.
+      for (size_t i = 0; i < components.size(); ++i)
+	{
+	  exp_type wanted = pe_unknown;
+
+	  switch (components[i].type)
+	    {
+
+	    case print_format::conv_unspecified:
+	    case print_format::conv_literal:
+	      assert (false);
+	      break;
+
+	    case print_format::conv_signed_decimal:
+	    case print_format::conv_unsigned_decimal:
+	    case print_format::conv_unsigned_octal:
+	    case print_format::conv_unsigned_uppercase_hex:
+	    case print_format::conv_unsigned_lowercase_hex:
+	      wanted = pe_long;
+	      break;
+
+	    case print_format::conv_string:
+	      wanted = pe_string;
+	      break;
+	    }
+
+	  assert (wanted != pe_unknown);
+
+	  t = wanted;
+	  e->args[i]->visit (this);
+
+	  if (e->args[i]->type == pe_unknown)
+	    {
+	      e->args[i]->type = wanted;
+	      resolved (e->args[i]->tok, wanted);
+	    }
+	  else if (e->args[i]->type != wanted)
+	    {
+	      mismatch (e->args[i]->tok, e->args[i]->type, wanted);
+	    }
+	}
+    }
+  else
+    {
+      // Without a format string, the best we can do is require that
+      // each argument resolve to a concrete type.
+      for (size_t i = 0; i < e->args.size(); ++i)
+	{
+	  t = pe_unknown;
+	  e->args[i]->visit (this);
+	  if (e->args[i]->type == pe_unknown)
+	    {
+	      unresolved (e->args[i]->tok);
+	      ++unresolved_args;
+	    }
+	}
+    }
+  
+  if (unresolved_args == 0)
+    {
+      if (e->type == pe_unknown)
+	{
+	  if (e->print_to_stream)
+	    e->type = pe_long;
+	  else
+	    e->type = pe_string;      
+	  resolved (e->tok, e->type);
+	}
+    }
+  else
+    {
+      e->type = pe_unknown;
+      unresolved (e->tok);
+    }
+}
+
+
+void 
+typeresolution_info::visit_stat_op (stat_op* e)
+{
+  t = pe_stats;
+  e->stat->visit (this);
+  if (e->type == pe_unknown)
+    {
+      e->type = pe_long;
+      resolved (e->tok, e->type);
+    }
+  else if (e->type != pe_long)
+    mismatch (e->tok, e->type, pe_long);
+}
+
+void 
+typeresolution_info::visit_hist_op (hist_op* e)
+{
+  t = pe_stats;
+  e->stat->visit (this);
+}
+
 
 void
 typeresolution_info::unresolved (const token* tok)
