@@ -9,6 +9,7 @@
 #ifndef ELABORATE_H
 #define ELABORATE_H
 
+#include "session.h"
 #include "staptree.h"
 #include "parse.h"
 #include <string>
@@ -19,7 +20,6 @@
 
 // ------------------------------------------------------------------------
 
-struct systemtap_session;
 struct derived_probe;
 struct match_node;
 
@@ -33,9 +33,6 @@ public:
   derived_probe* current_probe;
   symresolution_info (systemtap_session& s);
 
-  // XXX: instead in systemtap_session?
-  void derive_probes (match_node * root, probe *p, std::vector<derived_probe*>& dps);
-  
   vardecl* find_var (const std::string& name, int arity);
   functiondecl* find_function (const std::string& name, unsigned arity);
 
@@ -93,6 +90,9 @@ struct typeresolution_info: public visitor
   void visit_target_symbol (target_symbol* e);
   void visit_arrayindex (arrayindex* e);
   void visit_functioncall (functioncall* e);
+  void visit_print_format (print_format* e);
+  void visit_stat_op (stat_op* e);
+  void visit_hist_op (hist_op* e);
 };
 
 
@@ -143,7 +143,6 @@ struct derived_probe_builder
 		     probe* base, 
 		     probe_point* location,
 		     std::map<std::string, literal*> const & parameters,
-		     std::vector<probe*> & results_to_expand_further,
 		     std::vector<derived_probe*> & finished_results) = 0;
   virtual ~derived_probe_builder() {}
 
@@ -174,15 +173,18 @@ match_key
 class
 match_node
 {
-  std::map<match_key, match_node*> sub;
+  typedef std::map<match_key, match_node*> sub_map_t;
+  typedef std::map<match_key, match_node*>::iterator sub_map_iterator_t;
+  sub_map_t sub;
   derived_probe_builder* end;
 
  public:
   match_node();
-  derived_probe_builder* find_builder(std::vector<probe_point::component*> const & components,
-				       unsigned pos,
-				       std::vector< std::pair<std::string, literal*> > & parameters);
-  
+
+  void find_and_build (systemtap_session& s,
+                       probe* p, probe_point *loc, unsigned pos,
+                       std::vector<derived_probe *>& results);
+
   match_node* bind(match_key const & k);
   match_node* bind(std::string const & k);
   match_node* bind_str(std::string const & k);
@@ -192,61 +194,16 @@ match_node
 
 // ------------------------------------------------------------------------
 
-
-class unparser;
-class translator_output;
-
-struct systemtap_session
-{
-  systemtap_session ();
-
-  // command line args
-  std::vector<std::string> include_path;
-  std::vector<std::string> args;
-  std::string kernel_release;
-  std::string runtime_path;
-  std::string module_name;
-  std::string cmd;
-  int target_pid;
-  int last_pass;
-  bool test_mode;
-  bool verbose;
-  bool keep_tmpdir;
-  bool guru_mode;
-
-  // temporary directory for module builds etc.
-  // hazardous - it is "rm -rf"'d at exit
-  std::string tmpdir;
-  std::string translated_source; // C source code
-
-  match_node* pattern_root;
-  void register_library_aliases();
-
-  // parse trees for the various script files
-  stapfile* user_file;
-  std::vector<stapfile*> library_files;
-
-  // resolved globals/functions/probes for the run as a whole
-  std::vector<stapfile*> files;
-  std::vector<vardecl*> globals;
-  std::vector<functiondecl*> functions;
-  std::vector<derived_probe*> probes;
-  std::vector<embeddedcode*> embeds;
-
-  // module-referencing file handles
-  std::map<std::string,int> module_fds;
-
-  // unparser data
-  translator_output* op;
-  unparser* up;
-
-  unsigned num_errors;
-  // void print_error (const parse_error& e);
-  void print_error (const semantic_error& e);
-};
-
+/* struct systemtap_session moved to session.h */
 
 int semantic_pass (systemtap_session& s);
+void derive_probes (systemtap_session& s,
+                    probe *p, std::vector<derived_probe*>& dps,
+                    bool exc_outermost = true);
+
+// A helper we use here and in translate, for pulling symbols out of lvalue
+// expressions.
+symbol * get_symbol_within_expression (expression *e);
 
 
 #endif // ELABORATE_H
