@@ -39,9 +39,39 @@ fail (void *arg __attribute__ ((unused)), const char *fmt, ...)
   vfprintf (stderr, _(fmt), ap);
   va_end (ap);
 
+  fprintf (stderr, "\n");
+
   exit (2);
 }
 
+static const Dwarf_Op *
+get_location (Dwarf_Addr dwbias, Dwarf_Addr pc, Dwarf_Attribute *loc_attr,
+	      size_t *len)
+{
+  Dwarf_Op *expr;
+
+  switch (dwarf_getlocation_addr (loc_attr, pc - dwbias, &expr, len, 1))
+    {
+    case 1:			/* Should always happen.  */
+      if (*len == 0)
+	goto inaccessible;
+      break;
+
+    default:			/* Shouldn't happen.  */
+    case -1:
+      fail (NULL, _("dwarf_addrloclists (form %#x): %s"),
+	    dwarf_whatform (loc_attr), dwarf_errmsg (-1));
+      return NULL;
+
+    case 0:			/* Shouldn't happen.  */
+    inaccessible:
+      fail (NULL, _("not accessible at this address"));
+      return NULL;
+    }
+
+  return expr;
+}
+
 static void
 handle_variable (Dwarf_Die *scopes, int nscopes, int out,
 		 Dwarf_Addr cubias, Dwarf_Die *vardie, Dwarf_Addr pc,
@@ -84,9 +114,12 @@ handle_variable (Dwarf_Die *scopes, int nscopes, int out,
 #define FIELD "addr"
 #define emit(fmt, ...) printf ("  addr = " fmt "\n", ## __VA_ARGS__)
 
+  size_t locexpr_len;
+  const Dwarf_Op *locexpr = get_location (cubias, pc, &attr_mem, &locexpr_len);
+
   struct location *head, *tail = NULL;
   head = c_translate_location (&pool, &fail, NULL, NULL,
-			       1, cubias, &attr_mem, pc,
+			       1, cubias, pc, locexpr, locexpr_len,
 			       &tail, fb_attr);
 
   if (dwarf_attr_integrate (vardie, DW_AT_type, &attr_mem) == NULL)
@@ -97,8 +130,6 @@ handle_variable (Dwarf_Die *scopes, int nscopes, int out,
   Dwarf_Die die_mem, *die = vardie;
   while (*fields != NULL)
     {
-      die = dwarf_formref_die (&attr_mem, &die_mem);
-
       if (!strcmp (*fields, "="))
 	{
 	  store = true;
@@ -107,6 +138,8 @@ handle_variable (Dwarf_Die *scopes, int nscopes, int out,
 	  break;
 	}
 
+      die = dwarf_formref_die (&attr_mem, &die_mem);
+
       const int typetag = dwarf_tag (die);
       switch (typetag)
 	{
@@ -179,9 +212,12 @@ handle_variable (Dwarf_Die *scopes, int nscopes, int out,
 		       *fields, dwarf_errmsg (-1));
 	    }
 	  else
-	    c_translate_location (&pool, NULL, NULL, NULL,
-				  1, cubias, &attr_mem, pc,
-				  &tail, NULL);
+	    {
+	      locexpr = get_location (cubias, pc, &attr_mem, &locexpr_len);
+	      c_translate_location (&pool, NULL, NULL, NULL,
+				    1, cubias, pc, locexpr, locexpr_len,
+				    &tail, NULL);
+	    }
 	  ++fields;
 	  break;
 
@@ -258,6 +294,8 @@ handle_variable (Dwarf_Die *scopes, int nscopes, int out,
 
   bool deref = c_emit_location (stdout, head, 1);
 
+  obstack_free (&pool, NULL);
+
   puts (store ? " return;" :
 	"  printk (\" ---> %ld\\n\", (unsigned long) value);\n"
 	"  return;");
@@ -457,6 +495,8 @@ main (int argc, char **argv)
 	handle_variable (scopes, n, out, cubias, &vardie, pc, &argv[argi]);
     }
 
+  free (scopes);
+
   dwfl_end (dwfl);
 
   return 0;
