@@ -9,12 +9,14 @@
 #include "config.h"
 #include "staptree.h"
 #include "parse.h"
+#include "session.h"
 #include <iostream>
 #include <fstream>
 #include <cctype>
 #include <cstdlib>
 #include <cerrno>
 #include <climits>
+#include <sstream>
 
 using namespace std;
 
@@ -22,13 +24,15 @@ using namespace std;
 
 
 
-parser::parser (istream& i, bool p):
+parser::parser (systemtap_session& s, istream& i, bool p):
+  session (s),
   input_name ("<input>"), free_input (0),
   input (i, input_name), privileged (p),
   last_t (0), next_t (0), num_errors (0)
 { }
 
-parser::parser (const string& fn, bool p):
+parser::parser (systemtap_session& s, const string& fn, bool p):
+  session (s),
   input_name (fn), free_input (new ifstream (input_name.c_str(), ios::in)),
   input (* free_input, input_name), privileged (p),
   last_t (0), next_t (0), num_errors (0)
@@ -41,17 +45,17 @@ parser::~parser()
 
 
 stapfile*
-parser::parse (std::istream& i, bool pr)
+parser::parse (systemtap_session& s, std::istream& i, bool pr)
 {
-  parser p (i, pr);
+  parser p (s, i, pr);
   return p.parse ();
 }
 
 
 stapfile*
-parser::parse (const std::string& n, bool pr)
+parser::parse (systemtap_session& s, const std::string& n, bool pr)
 {
-  parser p (n, pr);
+  parser p (s, n, pr);
   return p.parse ();
 }
 
@@ -100,11 +104,18 @@ parser::print_error  (const parse_error &pe)
 {
   cerr << "parse error: " << pe.what () << endl;
 
-  const token* t = last_t;
-  if (t)
-    cerr << "\tsaw: " << *t << endl;
+  if (pe.tok)
+    {
+      cerr << "\tat: " << *pe.tok << endl;
+    }
   else
-    cerr << "\tsaw: " << input_name << " EOF" << endl;
+    {
+      const token* t = last_t;
+      if (t)
+        cerr << "\tsaw: " << *t << endl;
+      else
+        cerr << "\tsaw: " << input_name << " EOF" << endl;
+    }
 
   // XXX: make it possible to print the last input line,
   // so as to line up an arrow with the specific error column
@@ -120,11 +131,175 @@ parser::last ()
 }
 
 
+// Here, we perform on-the-fly preprocessing.
+// The basic form is %( CONDITION %? THEN-TOKENS %: ELSE-TOKENS %)
+// where CONDITION is: kernel_v[r] COMPARISON-OP "version-string"
+//                 or: arch COMPARISON-OP "arch-string"
+// The %: ELSE-TOKENS part is optional.
+//
+// e.g. %( kernel_v > "2.5" %? "foo" %: "baz" %)
+// e.g. %( arch != "i686" %? "foo" %: "baz" %)
+//
+// Up to an entire %( ... %) expression is processed by a single call
+// to this function.  Tokens included by any nested conditions are
+// enqueued in a private vector.
+
+bool eval_pp_conditional (systemtap_session& s,
+                          const token* l, const token* op, const token* r)
+{
+  if (l->type == tok_identifier && (l->content == "kernel_v" ||
+                                    l->content == "kernel_vr"))
+    {
+      string target_kernel_vr = s.kernel_release;
+      string target_kernel_v = target_kernel_vr;
+      // cut off any release code suffix
+      string::size_type dr = target_kernel_vr.rfind ('-');
+      if (dr > 0 && dr != string::npos)
+        target_kernel_v = target_kernel_vr.substr (0, dr);
+      
+      if (! (r->type == tok_string))
+        throw parse_error ("expected string literal", r);
+      string query_kernel_vr = r->content;
+      
+      // collect acceptable strverscmp results.
+      int rvc_ok1, rvc_ok2;
+      if (op->type == tok_operator && op->content == "<=")
+        { rvc_ok1 = -1; rvc_ok2 = 0; }
+      else if (op->type == tok_operator && op->content == ">=")
+        { rvc_ok1 = 1; rvc_ok2 = 0; }
+      else if (op->type == tok_operator && op->content == "<")
+        { rvc_ok1 = -1; rvc_ok2 = -1; }
+      else if (op->type == tok_operator && op->content == ">")
+        { rvc_ok1 = 1; rvc_ok2 = 1; }
+      else if (op->type == tok_operator && op->content == "==")
+        { rvc_ok1 = 0; rvc_ok2 = 0; }
+      else if (op->type == tok_operator && op->content == "!=")
+        { rvc_ok1 = -1; rvc_ok2 = 1; }
+      else
+        throw parse_error ("expected comparison operator", op);
+      
+      int rvc_result = strverscmp ((l->content == "kernel_vr" ? 
+                                    target_kernel_vr.c_str() :
+                                    target_kernel_v.c_str()),
+                                   query_kernel_vr.c_str());
+      // normalize rvc_result
+      if (rvc_result < 0) rvc_result = -1;
+      if (rvc_result > 0) rvc_result = 1;
+      
+      return (rvc_result == rvc_ok1 || rvc_result == rvc_ok2);
+    }
+  else if (l->type == tok_identifier && l->content == "arch")
+    {
+      string target_architecture = s.architecture;
+      if (! (r->type == tok_string))
+        throw parse_error ("expected string literal", r);
+      string query_architecture = r->content;
+      
+      bool result;
+      if (op->type == tok_operator && op->content == "==")
+        result = target_architecture == query_architecture;
+      else if (op->type == tok_operator && op->content == "!=")
+        result = target_architecture != query_architecture;
+      else
+        throw parse_error ("expected '==' or '!='", op);
+      
+      return result;
+    }  
+  // XXX: support other forms?  "CONFIG_SMP" ?
+  else
+    throw parse_error ("expected 'arch' or 'kernel_v' or 'kernel_vr'", l);
+}
+
+
+const token*
+parser::scan_pp ()
+{
+  while (true)
+    {
+      if (enqueued_pp.size() > 0)
+        {
+          const token* t = enqueued_pp[0];
+          enqueued_pp.erase (enqueued_pp.begin());
+          return t;
+        }
+
+      const token* t = input.scan (); // NB: not recursive!
+      if (t == 0) // EOF
+        return t;
+      
+      if (! (t->type == tok_operator && t->content == "%(")) // ordinary token
+        return t;
+
+      // We have a %( - it's time to throw a preprocessing party!
+
+      const token *l, *op, *r;
+      l = input.scan (); // NB: not recursive, though perhaps could be
+      op = input.scan ();
+      r = input.scan ();
+      if (l == 0 || op == 0 || r == 0)
+        throw parse_error ("incomplete condition after '%('", t);
+      // NB: consider generalizing to consume all tokens until %?, and
+      // passing that as a vector to an evaluator.
+
+      bool result = eval_pp_conditional (session, l, op, r);
+      
+      const token *m = input.scan (); // NB: not recursive
+      if (! (m && m->type == tok_operator && m->content == "%?"))
+        throw parse_error ("expected '%?' marker for conditional", t);
+
+      vector<const token*> my_enqueued_pp;
+      
+      while (true) // consume THEN tokens
+        {
+          m = scan_pp (); // NB: recursive
+          if (m == 0)
+            throw parse_error ("missing THEN tokens for conditional", t);
+          
+          if (m->type == tok_operator && (m->content == "%:" || // ELSE
+                                          m->content == "%)")) // END
+            break;
+          // enqueue token
+          if (result) 
+            my_enqueued_pp.push_back (m);
+          // continue
+        }
+      
+      if (m && m->type == tok_operator && m->content == "%:") // ELSE
+        while (true)
+          {
+            m = scan_pp (); // NB: recursive
+            if (m == 0)
+              throw parse_error ("missing ELSE tokens for conditional", t);
+            
+            if (m->type == tok_operator && m->content == "%)") // END
+              break;
+            // enqueue token
+            if (! result) 
+              my_enqueued_pp.push_back (m);
+            // continue
+          }
+
+      // NB: we transcribe the retained tokens here, and not inside
+      // the THEN/ELSE while loops.  If it were done there, each loop
+      // would become infinite (each iteration consuming an ordinary
+      // token the previous one just pushed there).  Guess how I
+      // figured that out.
+      enqueued_pp.insert (enqueued_pp.end(),
+                          my_enqueued_pp.begin(),
+                          my_enqueued_pp.end());
+
+      // Go back to outermost while(true) loop.  We hope that at least
+      // some THEN or ELSE tokens were enqueued.  If not, around we go
+      // again, until EOF.
+    }
+}
+
+
 const token*
 parser::next ()
 {
   if (! next_t)
-    next_t = input.scan ();
+    next_t = scan_pp ();
   if (! next_t)
     throw parse_error ("unexpected end-of-file");
 
@@ -139,9 +314,7 @@ const token*
 parser::peek ()
 {
   if (! next_t)
-    next_t = input.scan ();
-
-  // cerr << "{" << (next_t ? next_t->content : "null") << "}";
+    next_t = scan_pp ();
 
   // don't advance by zeroing next_t
   last_t = next_t;
@@ -160,7 +333,7 @@ const token*
 parser::expect_known (token_type tt, string const & expected)
 {
   const token *t = next();
-  if (! t && t->type == tt && t->content == expected)
+  if (! (t && t->type == tt && t->content == expected))
     throw parse_error ("expected '" + expected + "'");
   return t;
 }
@@ -190,6 +363,16 @@ parser::expect_kw (std::string const & expected)
   return expect_known (tok_identifier, expected);
 }
 
+const token* 
+parser::expect_number (int64_t & expected)
+{
+  std::string tmp;  
+  token const * tt = expect_unknown (tok_number, tmp);
+  istringstream iss(tmp);
+  iss >> expected;
+  return tt;
+}
+
 
 const token* 
 parser::expect_ident (std::string & target)
@@ -271,7 +454,7 @@ lexer::scan ()
   if (isspace (c))
     goto skip;
 
-  else if (isalpha (c) || c == '$' || c == '_')
+  else if (isalpha (c) || c == '$' || c == '@' || c == '_')
     {
       n->type = tok_identifier;
       n->content = (char) c;
@@ -463,7 +646,12 @@ lexer::scan ()
                s2 == "--" ||
                s2 == "->" ||
                s2 == "<<" ||
-               s2 == ">>")
+               s2 == ">>" ||
+               // preprocessor tokens
+               s2 == "%(" ||
+               s2 == "%?" ||
+               s2 == "%:" ||
+               s2 == "%)")
         {
           n->content = s2;
           input_get (); // swallow other character
@@ -518,17 +706,25 @@ parser::parse ()
       catch (parse_error& pe)
 	{
 	  print_error (pe);
-	  // Quietly swallow all tokens until the next '}'.
-	  while (1)
-	    {
-	      const token* t = peek ();
-	      if (! t)
-		break;
-	      next ();
-	      if (t->type == tok_operator && t->content == "}")
-		break;
-	    }
-	}
+          try 
+            {
+              // Quietly swallow all tokens until the next '}'.
+              while (1)
+                {
+                  const token* t = peek ();
+                  if (! t)
+                    break;
+                  next ();
+                  if (t->type == tok_operator && t->content == "}")
+                    break;
+                }
+            }
+          catch (parse_error& pe2)
+            {
+              // parse error during recovery ... ugh
+              print_error (pe2);
+            }
+        }
     }
 
   if (empty)
@@ -563,35 +759,28 @@ parser::parse_probe (std::vector<probe *> & probe_ret,
 
   while (1)
     {
-      const token *t = peek ();
-      if (t && t->type == tok_identifier)
-	{
-	  probe_point * pp = parse_probe_point ();
-
-	  t = peek ();
-	  if (equals_ok && t 
-	      && t->type == tok_operator && t->content == "=")
-	    {
-	      aliases.push_back(pp);
-	      next ();
-	      continue;
-	    }
-	  else if (t && t->type == tok_operator && t->content == ",")
-	    {
-	      locations.push_back(pp);
-	      equals_ok = false;
-	      next ();
-	      continue;
-	    }
-	  else if (t && t->type == tok_operator && t->content == "{")
-	    {
-	      locations.push_back(pp);
-	      break;
-	    }
-	  else
-            throw parse_error ("expected ',' or '{'");
-          // XXX: unify logic with that in parse_symbol()
-	}
+      probe_point * pp = parse_probe_point ();
+      
+      const token* t = peek ();
+      if (equals_ok && t 
+          && t->type == tok_operator && t->content == "=")
+        {
+          aliases.push_back(pp);
+          next ();
+          continue;
+        }
+      else if (t && t->type == tok_operator && t->content == ",")
+        {
+          locations.push_back(pp);
+          equals_ok = false;
+          next ();
+          continue;
+        }
+      else if (t && t->type == tok_operator && t->content == "{")
+        {
+          locations.push_back(pp);
+          break;
+        }
       else
 	throw parse_error ("expected probe point specifier");
     }
@@ -734,8 +923,8 @@ parser::parse_global (vector <vardecl*>& globals)
 
       for (unsigned i=0; i<globals.size(); i++)
 	if (globals[i]->name == t->content)
-          throw parse_error ("duplicate global name");
-
+	  throw parse_error ("duplicate global name");
+      
       vardecl* d = new vardecl;
       d->name = t->content;
       d->tok = t;
@@ -838,13 +1027,12 @@ parser::parse_probe_point ()
 {
   probe_point* pl = new probe_point;
 
-  // XXX: add support for probe point aliases
-  // e.g.   probe   alias = foo    { ... }
   while (1)
     {
       const token* t = next ();
-      if (t->type != tok_identifier)
-        throw parse_error ("expected identifier");
+      if (! (t->type == tok_identifier ||
+             (t->type == tok_operator && t->content == "*")))
+        throw parse_error ("expected identifier or '*'");
 
       if (pl->tok == 0) pl->tok = t;
 
@@ -946,6 +1134,8 @@ parser::parse_if_statement ()
       next ();
       s->elseblock = parse_statement ();
     }
+  else
+    s->elseblock = 0; // in case not otherwise initialized
 
   return s;
 }
@@ -1146,6 +1336,7 @@ parser::parse_foreach_loop ()
     throw parse_error ("expected 'foreach'");
   foreach_loop* s = new foreach_loop;
   s->tok = t;
+  s->sort_direction = 0;
 
   t = next ();
   if (! (t->type == tok_operator && t->content == "("))
@@ -1171,9 +1362,20 @@ parser::parse_foreach_loop ()
       sym->name = t->content;
       s->indexes.push_back (sym);
 
+      t = peek ();
+      if (t && t->type == tok_operator &&
+	  (t->content == "+" || t->content == "-"))
+	{
+	  if (s->sort_direction)
+	    throw parse_error ("multiple sort directives");
+	  s->sort_direction = (t->content == "+") ? 1 : -1;
+	  s->sort_column = s->indexes.size();
+	  next();
+	}
+
       if (parenthesized)
         {
-          const token* t = peek ();
+          t = peek ();
           if (t && t->type == tok_operator && t->content == ",")
             {
               next ();
@@ -1194,11 +1396,19 @@ parser::parse_foreach_loop ()
   t = next ();
   if (! (t->type == tok_identifier && t->content == "in"))
     throw parse_error ("expected 'in'");
+ 
+  s->base = parse_indexable();
 
-  t = next ();
-  if (t->type != tok_identifier)
-    throw parse_error ("expected identifier");
-  s->base = t->content;
+  t = peek ();
+  if (t && t->type == tok_operator &&
+      (t->content == "+" || t->content == "-"))
+    {
+      if (s->sort_direction)
+	throw parse_error ("multiple sort directives");
+      s->sort_direction = (t->content == "+") ? 1 : -1;
+      s->sort_column = 0;
+      next();
+    }
 
   t = next ();
   if (! (t->type == tok_operator && t->content == ")"))
@@ -1438,13 +1648,8 @@ parser::parse_array_in ()
 
       arrayindex* a = new arrayindex;
       a->indexes = indexes;
-
-      t = next ();
-      if (t->type != tok_identifier)
-        throw parse_error ("expected identifier");
-      a->tok = t;
-      a->base = t->content;
-
+      a->base = parse_indexable();
+      a->tok = a->base->get_tok();
       e->operand = a;
       return e;
     }
@@ -1656,50 +1861,244 @@ parser::parse_value ()
 }
 
 
-// var, var[index], func(parms), thread->var, process->var
+const token *
+parser::parse_hist_op_or_bare_name (hist_op *&hop, string &name)
+{
+  hop = NULL;
+  const token* t = expect_ident (name);
+  if (name == "@hist_linear" || name == "@hist_log")
+    {
+      hop = new hist_op;
+      if (name == "@hist_linear")
+	hop->htype = hist_linear;
+      else if (name == "@hist_log")
+	hop->htype = hist_log;
+      hop->tok = t;
+      expect_op("(");
+      hop->stat = parse_expression ();
+      int64_t tnum;
+      if (hop->htype == hist_linear)
+	{
+	  for (size_t i = 0; i < 3; ++i)
+	    {
+	      expect_op (",");
+	      expect_number (tnum);
+	      hop->params.push_back (tnum);
+	    }
+	}
+      else
+	{
+	  assert(hop->htype == hist_log);
+	  if (peek_op (","))
+	    {
+	      expect_op (",");
+	      expect_number (tnum);
+	      hop->params.push_back (tnum);
+	    }
+	  else
+	    {
+	      // FIXME (magic value): Logarithmic histograms get 64
+	      // buckets by default.
+	      hop->params.push_back (64);
+	    }	      
+	}
+      expect_op(")");
+    }
+  return t;
+}
+
+
+indexable*
+parser::parse_indexable ()
+{
+  hist_op *hop = NULL;
+  string name;
+  const token *tok = parse_hist_op_or_bare_name(hop, name);
+  if (hop)
+    return hop;
+  else
+    {
+      symbol* sym = new symbol;
+      sym->name = name;
+      sym->tok = tok;
+      return sym;
+    }
+}
+
+
+// var, indexable[index], func(parms), printf("...", ...), $var, $var->member, @stat_op(stat)
 expression*
 parser::parse_symbol () 
 {
+  hist_op *hop = NULL;
+  symbol *sym = NULL;
   string name;
-  const token* t = expect_ident (name);
-  const token* t2 = t;
-  
-  if (name.size() > 0 && name[0] == '$')
+  const token *t = parse_hist_op_or_bare_name(hop, name);
+
+  if (!hop)
     {
-      // target_symbol time
-      target_symbol *tsym = new target_symbol;
-      tsym->tok = t;
-      tsym->base_name = name;
-      while (true)
+      // If we didn't get a hist_op, then we did get an identifier. We can 
+      // now scrutinize this identifier for the various magic forms of identifier
+      // (printf, @stat_op, and $var...)
+
+      if (name.size() > 0 && name[0] == '@')
 	{
-	  string c;
-	  if (peek_op ("->"))
-	    { 
-	      next(); 
-	      expect_ident (c);
-	      tsym->components.push_back
-		(make_pair (target_symbol::comp_struct_member, c));
-	    }
-	  else if (peek_op ("["))
-	    { 
-	      next();
-	      expect_unknown (tok_number, c);
-	      expect_op ("]");
-	      tsym->components.push_back
-		(make_pair (target_symbol::comp_literal_array_index, c));
-	    }	    
+	  stat_op *sop = new stat_op;
+	  if (name == "@avg")
+	    sop->ctype = sc_average;
+	  else if (name == "@count")
+	    sop->ctype = sc_count;
+	  else if (name == "@sum")
+	    sop->ctype = sc_sum;
+	  else if (name == "@min")
+	    sop->ctype = sc_min;
+	  else if (name == "@max")
+	    sop->ctype = sc_max;
 	  else
-	    break;
+	    throw parse_error("unknown statistic operator " + name);
+	  expect_op("(");
+	  sop->tok = t;
+	  sop->stat = parse_expression ();
+	  expect_op(")");
+	  return sop;
+	}
+      
+      else if (name.size() > 0 && (name == "print"
+				   || name == "sprint"
+				   || name == "printf"
+				   || name == "sprintf"))
+	{
+	  print_format *fmt = new print_format;
+	  fmt->tok = t;
+	  fmt->print_with_format = (name[name.size() - 1] == 'f');
+	  fmt->print_to_stream = (name[0] == 'p');
+	  expect_op("(");
+	  if (fmt->print_with_format)
+	    {
+	      // Consume and convert a format string, and any subsequent
+	      // arguments. Agreement between the format string and the
+	      // arguments is postponed to the typechecking phase. 
+	      string tmp;
+	      expect_unknown (tok_string, tmp);
+	      fmt->components = print_format::string_to_components (tmp);
+	      while (!peek_op (")"))
+		{
+		  expect_op(",");
+		  expression *e = parse_expression ();	      
+		  fmt->args.push_back(e);
+		}
+	    }
+	  else if (name == "print" && 
+		   (peek_kw("@hist_linear") || 
+		    peek_kw("@hist_log")))
+	    {
+	      // We have a special case where we recognize
+	      // print(@hist_foo(bar)) as a magic print-the-histogram
+	      // construct. This is sort of gross but it avoids
+	      // promoting histogram references to typeful
+	      // expressions.
+	      fmt->hist = NULL;
+	      t = parse_hist_op_or_bare_name(fmt->hist, name);
+	      assert(fmt->hist);
+	    }
+	  else
+	    {
+	      // If we are not printing with a format string, we permit
+	      // exactly one argument (of any type).
+	      expression *e = parse_expression ();	      
+	      fmt->args.push_back(e);
+	    }
+	  expect_op(")");
+	  return fmt;
+	}
+      
+      else if (name.size() > 0 && name[0] == '$')
+	{
+	  // target_symbol time
+	  target_symbol *tsym = new target_symbol;
+	  tsym->tok = t;
+	  tsym->base_name = name;
+	  while (true)
+	    {
+	      string c;
+	      if (peek_op ("->"))
+		{ 
+		  next(); 
+		  expect_ident (c);
+		  tsym->components.push_back
+		    (make_pair (target_symbol::comp_struct_member, c));
+		}
+	      else if (peek_op ("["))
+		{ 
+		  next();
+		  expect_unknown (tok_number, c);
+		  expect_op ("]");
+		  tsym->components.push_back
+		    (make_pair (target_symbol::comp_literal_array_index, c));
+		}	    
+	      else
+		break;
+	    }
+	  return tsym;
+	}
+
+      else if (peek_op ("(")) // function call
+	{
+	  next ();
+	  struct functioncall* f = new functioncall;
+	  f->tok = t;
+	  f->function = name;
+	  // Allow empty actual parameter list
+	  if (peek_op (")"))
+	    {
+	      next ();
+	      return f;
+	    }
+	  while (1)
+	    {
+	      f->args.push_back (parse_expression ());
+	      if (peek_op (")"))
+		{
+		  next();
+		  break;
+		}
+	      else if (peek_op (","))
+		{
+		  next();
+		  continue;
+		}
+	      else
+		throw parse_error ("expected ',' or ')'");
+	    }
+	  return f;
+	}
+
+      else
+	{
+	  sym = new symbol;
+	  sym->name = name;
+	  sym->tok = t;
 	}
-      return tsym;
     }
   
+  // By now, either we had a hist_op in the first place, or else 
+  // we had a plain word and it was converted to a symbol.
+
+  assert (hop || sym);
+
+  // All that remains is to check for array indexing
+
   if (peek_op ("[")) // array
     {
       next ();
       struct arrayindex* ai = new arrayindex;
-      ai->tok = t2;
-      ai->base = name;
+      ai->tok = t;
+
+      if (hop)
+	ai->base = hop;
+      else
+	ai->base = sym;
+
       while (1)
         {
           ai->indexes.push_back (parse_expression ());
@@ -1718,42 +2117,14 @@ parser::parse_symbol ()
         }
       return ai;
     }
-  else if (peek_op ("(")) // function call
-    {
-      next ();
-      struct functioncall* f = new functioncall;
-      f->tok = t2;
-      f->function = name;
-      // Allow empty actual parameter list
-      if (peek_op (")"))
-	{
-	  next ();
-	  return f;
-	}
-      while (1)
-	{
-	  f->args.push_back (parse_expression ());
-	  if (peek_op (")"))
-	    {
-	      next();
-	      break;
-	    }
-	  else if (peek_op (","))
-	    {
-	      next();
-	      continue;
-	    }
-	  else
-	    throw parse_error ("expected ',' or ')'");
-	}
-      return f;
-    }
-  else
-    {
-      symbol* sym = new symbol;
-      sym->name = name;
-      sym->tok = t2;
-      return sym;
-    }
+
+  // If we got to here, we *should* have a symbol; if we have
+  // a hist_op on its own, it doesn't count as an expression,
+  // so we throw a parse error.
+
+  if (hop)
+    throw parse_error("base histogram operator where expression expected", t);
+  
+  return sym;  
 }
 
