@@ -1,4 +1,5 @@
-/* I/O for printing warnings, errors and debug messages
+/* -*- linux-c -*- 
+ * I/O for printing warnings, errors and debug messages
  * Copyright (C) 2005 Red Hat Inc.
  *
  * This file is part of systemtap, and is free software.  You can
@@ -24,8 +25,8 @@ void _stp_string_cat_cstr (String str1, const char *str2);
 
 /** private buffer for _stp_log() */
 #define STP_LOG_BUF_LEN 2047
-#define WARN_STRING "\033[33mWARNING: \033[0m"
-#define ERR_STRING "\033[31mERROR: \033[0m"
+#define WARN_STRING "WARNING: "
+#define ERR_STRING "ERROR: "
 
 static char _stp_lbuf[NR_CPUS][STP_LOG_BUF_LEN + 1];
 
@@ -115,7 +116,7 @@ void _stp_warn (const char *fmt, ...)
  */
 void _stp_exit (void)
 {
-	schedule_work (&stp_exit);
+	_stp_exit_flag = 1;
 }
 
 /** Prints error message and exits.
