@@ -1,4 +1,12 @@
-/* target operations */
+/* target operations
+ * Copyright (C) 2005 Red Hat Inc.
+ * Copyright (C) 2005 Intel Corporation.
+ *
+ * This file is part of systemtap, and is free software.  You can
+ * redistribute it and/or modify it under the terms of the GNU General
+ * Public License (GPL); either version 2, or (at your option) any
+ * later version.
+ */
 
 #include <linux/types.h>
 #define intptr_t long
@@ -50,6 +58,13 @@
 #define dwarf_register_6(regs)	regs->esi
 #define dwarf_register_7(regs)	regs->edi
 
+#elif defined __ia64__
+#undef fetch_register
+#undef store_register
+
+#define fetch_register(regno)		ia64_fetch_register(regno, c->regs)
+#define store_register(regno, value)	ia64_store_register(regno, c->regs, value)
+
 #elif defined __x86_64__
 
 #define dwarf_register_0(regs)	regs->rax
@@ -72,6 +87,7 @@
 #elif defined __powerpc__
 
 #undef fetch_register
+#undef store_register
 #define fetch_register(regno) ((intptr_t) c->regs->gpr[regno])
 #define store_register(regno) (c->regs->gpr[regno] = (value))
 
@@ -102,16 +118,47 @@
     int _bad = 0;							      \
     switch (size)							      \
       {									      \
-      case 1: __put_user_asm(((u8)(value),addr,_bad,"b","b","iq",1); break;   \
-      case 2: __put_user_asm(((u16)(value),addr,_bad,"w","w","ir",1); break;  \
-      case 4: __put_user_asm(((u32)(value),addr,_bad,"l","k","ir",1); break;  \
-      case 8: __put_user_asm(((u64)(value),addr,_bad,"q","","ir",1); break;   \
+      case 1: __put_user_asm(((u8)(value)),addr,_bad,"b","b","iq",1); break;  \
+      case 2: __put_user_asm(((u16)(value)),addr,_bad,"w","w","ir",1); break; \
+      case 4: __put_user_asm(((u32)(value)),addr,_bad,"l","k","ir",1); break; \
+      case 8: __put_user_asm(((u64)(value)),addr,_bad,"q","","ir",1); break;  \
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
       goto deref_fault;							      \
   })
 
+#elif defined __ia64__								
+#define deref(size, addr)						\
+  ({									\
+     int _bad = 0;							\
+     intptr_t _v=0;							\
+	switch (size){							\
+	case 1: __get_user_size(_v, addr, 1, _bad); break; 		\
+	case 2: __get_user_size(_v, addr, 2, _bad); break;  		\
+	case 4: __get_user_size(_v, addr, 4, _bad); break;  		\
+	case 8: __get_user_size(_v, addr, 8, _bad); break;  		\
+	default: __get_user_unknown(); break;				\
+	}								\
+    if (_bad)  								\
+	goto deref_fault;						\
+     _v;								\
+   })
+
+#define store_deref(size, addr, value)					\
+  ({									\
+    int _bad=0;								\
+	switch (size){							\
+	case 1: __put_user_size(value, addr, 1, _bad); break;		\
+	case 2: __put_user_size(value, addr, 2, _bad); break;		\
+	case 4: __put_user_size(value, addr, 4, _bad); break;		\
+	case 8: __put_user_size(value, addr, 8, _bad); break;		\
+	default: __put_user_unknown(); break;				\
+	}								\
+    if (_bad)								\
+	   goto deref_fault;						\
+   })
+
 #elif defined __powerpc64__
 
 #define deref(size, addr)						      \
@@ -136,10 +183,10 @@
     int _bad = 0;							      \
     switch (size)							      \
       {									      \
-      case 1: __put_user_asm(((u8)(value),addr,_bad,"stb",1); break;   	      \
-      case 2: __put_user_asm(((u16)(value),addr,_bad,"sth",1); break;  	      \
-      case 4: __put_user_asm(((u32)(value),addr,_bad,"stw",1); break;  	      \
-      case 8: __put_user_asm(((u64)(value),addr,_bad,"std",1); break; 	      \
+      case 1: __put_user_asm(((u8)(value)),addr,_bad,"stb",1); break;	      \
+      case 2: __put_user_asm(((u16)(value)),addr,_bad,"sth",1); break;	      \
+      case 4: __put_user_asm(((u32)(value)),addr,_bad,"stw",1); break;	      \
+      case 8: __put_user_asm(((u64)(value)),addr,_bad,"std",1); break;	      \
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
@@ -170,10 +217,10 @@
     int _bad = 0;							      \
     switch (size)							      \
       {									      \
-      case 1: __put_user_asm(((u8)(value),addr,_bad,"stb"); break;   	      \
-      case 2: __put_user_asm(((u16)(value),addr,_bad,"sth"); break;  	      \
-      case 4: __put_user_asm(((u32)(value),addr,_bad,"stw"); break;  	      \
-      case 8: __put_user_asm2(((u64)(value),addr,_bad); break;		      \
+      case 1: __put_user_asm(((u8)(value)),addr,_bad,"stb"); break;	      \
+      case 2: __put_user_asm(((u16)(value)),addr,_bad,"sth"); break;	      \
+      case 4: __put_user_asm(((u32)(value)),addr,_bad,"stw"); break;	      \
+      case 8: __put_user_asm2(((u64)(value)),addr,_bad); break;		      \
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
