@@ -9,42 +9,15 @@
  */
 
 /** @file map-stat.c
- * @brief Map functions to set and get stats
+ * @brief Map functions to handle statistics.
  */
 
 #include "stat-common.c"
 
-/* Adds an int64 to a stats map */
-static void _stp_map_add_stat (MAP map, int64_t val)
-{
-	stat *d;
-	Stat st;
-
-	if (map == NULL)
-		return;
-		
-	if (map->create) {
-		struct map_node *m = __stp_map_create (map);
-		if (!m)
-			return;
-		
-		/* set the value */
-		d = (stat *)((long)m + map->data_offset);
-		d->count = 0;
-	} else {
-		if (map->key == NULL)
-			return;
-		d = (stat *)((long)map->key + map->data_offset);
-	}
-	st = (Stat)((long)map + offsetof(struct map_root, hist_type));
-	__stp_stat_add (st, d, val);
-}
-
 
 static void _stp_map_print_histogram (MAP map, stat *sd)
 {
-	Stat st = (Stat)((long)map + offsetof(struct map_root, hist_type));
-	_stp_stat_print_histogram (st, sd);
+	_stp_stat_print_histogram (&map->hist, sd);
 }
 
 static MAP _stp_map_new_hstat_log (unsigned max_entries, int key_size, int buckets)
@@ -53,11 +26,11 @@ static MAP _stp_map_new_hstat_log (unsigned max_entries, int key_size, int bucke
 	int size = buckets * sizeof(int64_t) + sizeof(stat);
 	MAP m = _stp_map_new (max_entries, STAT, key_size, size);
 	if (m) {
-		m->hist_type = HIST_LOG;
-		m->hist_buckets = buckets;
+		m->hist.type = HIST_LOG;
+		m->hist.buckets = buckets;
 		if (buckets < 1 || buckets > 64) {
 			_stp_warn("histogram: Bad number of buckets.  Must be between 1 and 64\n");
-			m->hist_type = HIST_NONE;
+			m->hist.type = HIST_NONE;
 			return m;
 		}
 	}
@@ -76,17 +49,71 @@ static MAP _stp_map_new_hstat_linear (unsigned max_entries, int ksize, int start
 
 	m = _stp_map_new (max_entries, STAT, ksize, size);
 	if (m) {
-		m->hist_type = HIST_LINEAR;
-		m->hist_start = start;
-		m->hist_stop = stop;
-		m->hist_int = interval;
-		m->hist_buckets = buckets;
-		if (m->hist_buckets <= 0) {
+		m->hist.type = HIST_LINEAR;
+		m->hist.start = start;
+		m->hist.stop = stop;
+		m->hist.interval = interval;
+		m->hist.buckets = buckets;
+		if (m->hist.buckets <= 0) {
 			_stp_warn("histogram: bad stop, start and/or interval\n");
-			m->hist_type = HIST_NONE;
+			m->hist.type = HIST_NONE;
 			return m;
 		}
 		
 	}
 	return m;
 }
+
+static MAP _stp_pmap_new_hstat_linear (unsigned max_entries, int ksize, int start, int stop, int interval)
+{
+	MAP map;
+	int size;
+	int buckets = (stop - start) / interval;
+	if ((stop - start) % interval) buckets++;
+
+        /* add size for buckets */
+	size = buckets * sizeof(int64_t) + sizeof(stat);
+
+	map = _stp_pmap_new (max_entries, STAT, ksize, size);
+	if (map) {
+		int i;
+		MAP m;
+		for_each_cpu(i) {
+			m = _stp_per_cpu_ptr (map, i);
+			m->hist.type = HIST_LINEAR;
+			m->hist.start = start;
+			m->hist.stop = stop;
+			m->hist.interval = interval;
+			m->hist.buckets = buckets;
+		}
+		/* now set agg map  params */
+		m = _stp_percpu_dptr(map);
+		m->hist.type = HIST_LINEAR;
+		m->hist.start = start;
+		m->hist.stop = stop;
+		m->hist.interval = interval;
+		m->hist.buckets = buckets;
+	}
+	return map;
+}
+
+static MAP _stp_pmap_new_hstat_log (unsigned max_entries, int key_size, int buckets)
+{
+	/* add size for buckets */
+	int size = buckets * sizeof(int64_t) + sizeof(stat);
+	MAP map = _stp_pmap_new (max_entries, STAT, key_size, size);
+	if (map) {
+		int i;
+		MAP m;
+		for_each_cpu(i) {
+			m = _stp_per_cpu_ptr (map, i);
+			m->hist.type = HIST_LOG;
+			m->hist.buckets = buckets;
+		}
+		/* now set agg map  params */
+		m = _stp_percpu_dptr(map);
+		m->hist.type = HIST_LOG;
+		m->hist.buckets = buckets;
+	}
+	return map;
+}
