@@ -1,4 +1,14 @@
-#ifndef _MAP_H_ /* -*- linux-c -*- */
+/* -*- linux-c -*- 
+ * Map Header File
+ * Copyright (C) 2005 Red Hat Inc.
+ *
+ * This file is part of systemtap, and is free software.  You can
+ * redistribute it and/or modify it under the terms of the GNU General
+ * Public License (GPL); either version 2, or (at your option) any
+ * later version.
+ */
+
+#ifndef _MAP_H_
 #define _MAP_H_
 
 /** @file map.h
@@ -37,11 +47,6 @@ amount of memory. Do not increase above 5. */
 
 #include "stat.h"
 
-/* Histogram is log 2 */
-#define HSTAT_LOG (STAT | (HIST_LOG << 8))
-/* Histogram is linear */
-#define HSTAT_LINEAR (STAT | (HIST_LINEAR <<  8))
-
 /* Keys are either int64 or strings */
 typedef union {
 	int64_t val;
@@ -73,7 +78,10 @@ struct map_root {
 	int num;
 
 	/* when more than maxnum elements, wrap or discard? */
-	int no_wrap;
+	int wrap;
+
+	/* is this a list? */
+	int list;
 
 	/* linked list of current entries */
 	struct list_head head;
@@ -89,13 +97,10 @@ struct map_root {
 	/* space. */
 	void (*copy_keys)(struct map_root *, struct map_node *);
 	key_data (*get_key)(struct map_node *mn, int n, int *type);
-	int data_offset;
+	void (*copy)(struct map_node *dst, struct map_node *src);
+	int (*cmp)(struct map_node *dst, struct map_node *src);
 
-	/* this is the creation data saved between the key functions and the
-	    set/get functions */
-	u_int8_t create;
-	key_data c_key[MAX_KEY_ARITY];
-	struct hlist_head *c_keyhead;
+	int data_offset;
 
 	/* the hash table for this array */
 	struct hlist_head hashes[HASH_TABLE_SIZE];
@@ -104,11 +109,7 @@ struct map_root {
 	void *membuf;
 
 	/* used if this map's nodes contain stats */
-	enum histtype hist_type;
-	int hist_start;
-	int hist_stop;
-	int hist_int;
-	int hist_buckets;
+	struct _Hist hist;
 };
 
 /** All maps are of this type. */
@@ -123,47 +124,6 @@ typedef struct map_root *MAP;
 /** Extracts int from key2 union */
 #define key2int(ptr) (_stp_key_get_int64(ptr,2))
 
-/** Macro to call the proper _stp_map_key functions based on the
- * types of the arguments. 
- */
-#define _stp_map_key2(map, key1, key2)				\
-({								\
-	if (__builtin_types_compatible_p (typeof (key1), char[]))	\
-		if (__builtin_types_compatible_p (typeof (key2), char[])) \
-			_stp_map_key_str_str (map, (char *)(key1), (char *)(key2)); \
-		else							\
-			_stp_map_key_str_int64 (map, (char *)(key1), (int64_t)(key2)); \
-	else								\
-		if (__builtin_types_compatible_p (typeof (key2), char[])) \
-			_stp_map_key_int64_str (map, (int64_t)(key1), (char *)(key2)); \
-		else							\
-			_stp_map_key_int64_int64 (map, (int64_t)(key1), (int64_t)(key2)); \
-})
-
-/** Macro to call the proper _stp_map_key function based on the
- * type of the argument. 
- */
-#define _stp_map_key(map, key)				\
-  ({								\
-    if (__builtin_types_compatible_p (typeof (key), char[]))	\
-      _stp_map_key_str (map, (char *)(key));				\
-    else							\
-      _stp_map_key_int64 (map, (int64_t)(key));				\
-  })
-
-/** Macro to call the proper _stp_map_set function based on the
- * type of the argument. 
- * @ingroup map_set
- */
-#define _stp_map_set(map, val)					\
-  ({								\
-    if (__builtin_types_compatible_p (typeof (val), char[]))		\
-      _stp_map_set_str (map, (char *)(val));				\
-    else  if (__builtin_types_compatible_p (typeof (val), String))	\
-      _stp_map_set_string (map, (String)(val));				\
-    else								\
-      _stp_map_set_int64 (map, (int64_t)(val));				\
-  })
 
 /** Loop through all elements of a map or list.
  * @param map 
@@ -178,23 +138,6 @@ typedef struct map_root *MAP;
 
 /** @} */
 
-/** @ingroup lists
- * @brief Macro to call the proper _stp_list_add function based on the
- * types of the argument. 
- *
- * @note May cause compiler warning on some GCCs 
- */
-
-#define _stp_list_add(map, val)					\
-  ({									\
-    if (__builtin_types_compatible_p (typeof (val), char[]))		\
-      _stp_list_add_str (map, (char *)(val));				\
-    else if (__builtin_types_compatible_p (typeof (val), String))	\
-      _stp_list_add_string (map, (String)(val));			\
-    else								\
-      _stp_list_add_int64 (map, (int64_t)(val));			\
-  })
-
 
 /** @cond DONT_INCLUDE */
 /************* prototypes for map.c ****************/
@@ -205,7 +148,7 @@ void int64_add(void *dest, int64_t val);
 int64_t int64_get(void *ptr);
 void stat_copy(void *dest, stat *src);
 void stat_add(void *dest, stat *src);
-stat * stat_get(void *ptr);
+stat *stat_get(void *ptr);
 int64_t _stp_key_get_int64(struct map_node *mn, int n);
 char * _stp_key_get_str(struct map_node *mn, int n);
 unsigned int int64_hash(const int64_t v);
@@ -215,33 +158,32 @@ void str_add(void *dest, char *val);
 int str_eq_p(char *key1, char *key2);
 int64_t _stp_get_int64(struct map_node *m);
 char * _stp_get_str(struct map_node *m);
-stat * _stp_get_stat(struct map_node *m);
+stat *_stp_get_stat(struct map_node *m);
 unsigned int str_hash(const char *key1);
 static MAP _stp_map_new(unsigned max_entries, int type, int key_size, int data_size);
-#ifdef NEED_STAT_VALS
+static MAP _stp_pmap_new(unsigned max_entries, int type, int key_size, int data_size);
 static int msb64(int64_t x);
 static MAP _stp_map_new_hstat_log(unsigned max_entries, int key_size, int buckets);
 static MAP _stp_map_new_hstat_linear(unsigned max_entries, int ksize, int start, int stop, int interval);
 static void _stp_map_print_histogram(MAP map, stat *s);
-#endif
-void _stp_map_key_del(MAP map);
 struct map_node * _stp_map_start(MAP map);
 struct map_node * _stp_map_iter(MAP map, struct map_node *m);
 void _stp_map_del(MAP map);
+void _stp_map_clear(MAP map);
 void _stp_map_print(MAP map, const char *fmt);
-static struct map_node * __stp_map_create(MAP map);
 
-/* these prototypes suppress warnings from macros */
-void _stp_map_key_str(MAP, char *);
-void _stp_map_set_str(MAP, char *);
-void _stp_map_set_string(MAP, String);
-void _stp_list_add_str(MAP, char*);
-void _stp_list_add_string(MAP, String);
-
-void _stp_map_key_int64(MAP, int64_t);
-void _stp_map_set_int64(MAP, int64_t);
-int64_t _stp_map_get_int64(MAP);
-
-unsigned _stp_map_entry_exists(MAP);
+static struct map_node *_new_map_create (MAP map, struct hlist_head *head);
+static int _new_map_set_int64 (MAP map, struct map_node *n, int64_t val, int add);
+static int64_t _new_map_get_int64 (MAP map, struct map_node *n);
+static char *_new_map_get_str (MAP map, struct map_node *n);
+static int _new_map_set_str (MAP map, struct map_node *n, char *val, int add);
+static stat *_new_map_get_stat (MAP map, struct map_node *n);
+static int _new_map_set_stat (MAP map, struct map_node *n, int64_t val, int add);
+static void _new_map_clear_node (struct map_node *);
+static void _new_map_del_node (MAP map, struct map_node *n);
+static MAP _stp_pmap_new_hstat_linear (unsigned max_entries, int ksize, int start, int stop, int interval);
+static MAP _stp_pmap_new_hstat_log (unsigned max_entries, int key_size, int buckets);
+static void _stp_add_agg(struct map_node *aptr, struct map_node *ptr);
+static struct map_node *_stp_new_agg(MAP agg, struct hlist_head *ahead, struct map_node *ptr);
 /** @endcond */
 #endif /* _MAP_H_ */
