@@ -1,4 +1,14 @@
-#ifndef _PRINT_C_ /* -*- linux-c -*- */
+/* -*- linux-c -*- 
+ * Print Functions
+ * Copyright (C) 2005 Red Hat Inc.
+ *
+ * This file is part of systemtap, and is free software.  You can
+ * redistribute it and/or modify it under the terms of the GNU General
+ * Public License (GPL); either version 2, or (at your option) any
+ * later version.
+ */
+
+#ifndef _PRINT_C_
 #define _PRINT_C_
 
 #include <linux/config.h>
@@ -61,12 +71,12 @@ void _stp_print_flush (void)
 #else /* STP_RELAYFS */
 
 /* size of timestamp, in bytes, including space */
-#define TIMESTAMP_SIZE 11
+#define TIMESTAMP_SIZE (sizeof(int))
 #define STP_PRINT_BUF_START (TIMESTAMP_SIZE)
 
 /** Size of buffer, not including terminating NULL */
 #ifndef STP_PRINT_BUF_LEN
-#define STP_PRINT_BUF_LEN (8192 - TIMESTAMP_SIZE - 2)
+#define STP_PRINT_BUF_LEN (8192 - TIMESTAMP_SIZE - 1)
 #endif
 
 static char _stp_pbuf[NR_CPUS][STP_PRINT_BUF_LEN + STP_PRINT_BUF_START + 1];
@@ -82,14 +92,12 @@ void _stp_print_flush (void)
 	int cpu = smp_processor_id();
 	char *buf = &_stp_pbuf[cpu][0];
 	char *ptr = buf + STP_PRINT_BUF_START;
-	int ret, seq;
+	int ret;
 
 	if (_stp_pbuf_len[cpu] == 0)
 		return;
 
-	seq = _stp_seq_inc();
-	scnprintf (buf, TIMESTAMP_SIZE, "%10d", seq);
-	buf[TIMESTAMP_SIZE - 1] = ' ';
+	*((int *)buf) = _stp_seq_inc();
 	ret = _stp_transport_write(buf, _stp_pbuf_len[cpu] + TIMESTAMP_SIZE + 1);
 	if (unlikely(ret < 0)) {
 #if 0
