@@ -1,3 +1,13 @@
+/* Stack tracing functions
+ * Copyright (C) 2005 Red Hat Inc.
+ * Copyright (C) 2005 Intel Corporation.
+ *
+ * This file is part of systemtap, and is free software.  You can
+ * redistribute it and/or modify it under the terms of the GNU General
+ * Public License (GPL); either version 2, or (at your option) any
+ * later version.
+ */
+
 #ifndef _STACK_C_ /* -*- linux-c -*- */
 #define _STACK_C_
 
@@ -38,6 +48,62 @@ static void __stp_stack_sprint (String str, unsigned long *stack, int verbose, i
 	}
 }
 
+#elif defined (__ia64__)
+struct dump_para{
+  unsigned long *sp;
+  String str;
+};
+
+static void __stp_show_stack_sym(struct unw_frame_info *info, void *arg)
+{
+   unsigned long ip, skip=1;
+   String str = ((struct dump_para*)arg)->str;
+   struct pt_regs *regs = container_of(((struct dump_para*)arg)->sp, struct pt_regs, r12);
+
+	do {
+		unw_get_ip(info, &ip);
+		if (ip == 0) break;
+                if (skip){
+			if (ip == REG_IP(regs))
+				skip = 0;
+                        else continue;
+                }
+		_stp_string_cat(str, " ");
+		_stp_symbol_sprint(str, ip);
+		_stp_string_cat (str, "\n");
+        } while (unw_unwind(info) >= 0);
+}
+
+static void __stp_show_stack_addr(struct unw_frame_info *info, void *arg)
+{
+   unsigned long ip, skip=1;
+   String str = ((struct dump_para*)arg)->str;
+   struct pt_regs *regs = container_of(((struct dump_para*)arg)->sp, struct pt_regs, r12);	
+
+	do {
+		unw_get_ip(info, &ip);
+		if (ip == 0) break;
+		if (skip){
+			if (ip == REG_IP(regs))
+				skip = 0;
+			continue;
+		}
+		_stp_sprintf (str, "%lx ", ip);
+	} while (unw_unwind(info) >= 0);
+}
+
+static void __stp_stack_sprint (String str, unsigned long *stack, int verbose, int levels)
+{
+  struct dump_para para;
+
+	para.str = str;
+	para.sp  = stack; 
+	if (verbose)
+	    unw_init_running(__stp_show_stack_sym, &para);
+        else
+	    unw_init_running(__stp_show_stack_addr, &para);
+}
+
 #elif  defined (__i386__)
 
 static inline int valid_stack_ptr(struct thread_info *tinfo, void *p)
@@ -95,7 +161,61 @@ static void __stp_stack_sprint (String str, unsigned long *stack, int verbose, i
 			break;
 	}
 }
+#elif defined (__powerpc64__)
 
+static void __stp_stack_sprint (String str, unsigned long *_sp,
+				int verbose, int levels)
+{
+	unsigned long ip, newsp, lr = 0;
+	int count = 0;
+	unsigned long sp = (unsigned long)_sp;
+	int firstframe = 1;
+
+	lr = 0;
+	do {
+		if (sp < KERNELBASE)
+			return;
+		_sp = (unsigned long *) sp;
+		newsp = _sp[0];
+		ip = _sp[2];
+		if (!firstframe || ip != lr) {
+			if (verbose) {
+				_stp_sprintf(str, "[%016lx] [%016lx] ", sp, ip);
+				_stp_symbol_sprint(str, ip);
+				if (firstframe)
+					_stp_string_cat(str, " (unreliable)");
+			}
+			else
+				_stp_sprintf(str,"%lx ", ip);
+		}
+		firstframe = 0;
+		/*
+		 * See if this is an exception frame.
+		 * We look for the "regshere" marker in the current frame.
+		 */
+		if ( _sp[12] == 0x7265677368657265ul) {
+			struct pt_regs *regs = (struct pt_regs *)
+				(sp + STACK_FRAME_OVERHEAD);
+			if (verbose) {
+				_stp_sprintf(str, "--- Exception: %lx at ",
+						regs->trap);
+				_stp_symbol_sprint(str, regs->nip);
+				_stp_string_cat(str, "\n");
+				lr = regs->link;
+				_stp_string_cat(str, "    LR =");
+				_stp_symbol_sprint(str, lr);
+				_stp_string_cat(str, "\n");
+				firstframe = 1;
+			}
+			else {
+				_stp_sprintf(str, "%lx ",regs->nip);
+				_stp_sprintf(str, "%lx ",regs->link);
+			}
+		}
+
+		sp = newsp;
+	} while (str->len < STP_STRING_SIZE);
+}
 
 #else
 #error "Unsupported architecture"
