@@ -1,8 +1,15 @@
-#ifndef _STAT_COMMON_C_ /* -*- linux-c -*- */
+/* -*- linux-c -*- 
+ * common stats functions for aggragations and maps
+ * Copyright (C) 2005 Red Hat Inc.
+ *
+ * This file is part of systemtap, and is free software.  You can
+ * redistribute it and/or modify it under the terms of the GNU General
+ * Public License (GPL); either version 2, or (at your option) any
+ * later version.
+ */
+
+#ifndef _STAT_COMMON_C_
 #define _STAT_COMMON_C_
-
-/* common stats functions for aggragations and maps */
-
 #include "stat.h"
 
 static int needed_space(int64_t v)
@@ -89,19 +96,36 @@ static int msb64(int64_t val)
 #define HIST_WIDTH 50
 #endif
 
-static void _stp_stat_print_histogram (Stat st, stat *sd)
+static void _stp_stat_print_histogram (Hist st, stat *sd)
 {
-	int scale, i, j, val_space, cnt_space;
+	int scale, i, j, val_space, cnt_space, 
+		low_bucket = -1, high_bucket = 0;
 	int64_t val, v, max = 0;
 
-	if (st->hist_type != HIST_LOG && st->hist_type != HIST_LINEAR)
+	if (st->type != HIST_LOG && st->type != HIST_LINEAR)
 		return;
-	
-	/* get the maximum value, for scaling */
-	for (i = 0; i < st->hist_buckets; i++)
+
+	/* Get the maximum value, for scaling. Also calculate the low
+	   and high values to bound the reporting range. */
+	for (i = 0; i < st->buckets; i++) {
+		if (sd->histogram[i] > 0 && low_bucket == -1)
+			low_bucket = i;
+		if (sd->histogram[i] > 0)
+			high_bucket = i;
 		if (sd->histogram[i] > max)
 			max = sd->histogram[i];
-	
+	}
+
+	/* Touch up the bucket margin to show up to two zero-slots on
+	   either side of the data range, seems aesthetically pleasant. */
+	for (i = 0; i < 2; i++) {
+		if (low_bucket > 0)
+			low_bucket--;
+		
+		if (high_bucket < (st->buckets-1))
+			high_bucket++;
+	}
+
 	if (max <= HIST_WIDTH)
 		scale = 1;
 	else {
@@ -112,10 +136,10 @@ static void _stp_stat_print_histogram (Stat st, stat *sd)
 	}
 
 	cnt_space = needed_space (max);
-	if (st->hist_type == HIST_LINEAR)
-		val_space = needed_space (st->hist_start +  st->hist_int * (st->hist_buckets - 1));
+	if (st->type == HIST_LINEAR)
+		val_space = needed_space (st->start +  st->interval * high_bucket);
 	else
-		val_space = needed_space (1 << (st->hist_buckets - 1));
+		val_space = needed_space (((int64_t)1) << high_bucket);
 	//dbug ("max=%lld scale=%d val_space=%d\n", max, scale, val_space);
 
 	/* print header */
@@ -130,24 +154,26 @@ static void _stp_stat_print_histogram (Stat st, stat *sd)
 	reprint (HIST_WIDTH, "-");
 	_stp_print_cstr (" count\n");
 	_stp_print_flush();
-	if (st->hist_type == HIST_LINEAR)
-		val = st->hist_start;
+	if (st->type == HIST_LINEAR)
+		val = st->start;
 	else
 		val = 0;
-	for (i = 0; i < st->hist_buckets; i++) {
-		reprint (val_space - needed_space(val), " ");
-		_stp_printf("%d", val);
-		_stp_print_cstr (" |");
-
-		/* v = s->histogram[i] / scale; */
-		v = sd->histogram[i];
-		do_div (v, scale);
+	for (i = 0; i < st->buckets; i++) {
+		if (i >= low_bucket && i <= high_bucket) {
+			reprint (val_space - needed_space(val), " ");
+			_stp_printf("%lld", val);
+			_stp_print_cstr (" |");
+			
+			/* v = s->histogram[i] / scale; */
+			v = sd->histogram[i];
+			do_div (v, scale);
 		
-		reprint (v, "@");
-		reprint (HIST_WIDTH - v + 1 + cnt_space - needed_space(sd->histogram[i]), " ");
-		_stp_printf ("%lld\n", sd->histogram[i]);
-		if (st->hist_type == HIST_LINEAR) 
-			val += st->hist_int;
+			reprint (v, "@");
+			reprint (HIST_WIDTH - v + 1 + cnt_space - needed_space(sd->histogram[i]), " ");
+			_stp_printf ("%lld\n", sd->histogram[i]);
+		}
+		if (st->type == HIST_LINEAR) 
+			val += st->interval;
 		else if (val == 0)
 			val = 1;
 		else
@@ -155,7 +181,7 @@ static void _stp_stat_print_histogram (Stat st, stat *sd)
 	}
 }
 
-static void _stp_stat_print_valtype (char *fmt, Stat st, struct stat_data *sd, int cpu)
+static void _stp_stat_print_valtype (char *fmt, Hist st, stat *sd, int cpu)
 {
 	switch (*fmt) {
 	case 'C':
@@ -173,10 +199,8 @@ static void _stp_stat_print_valtype (char *fmt, Stat st, struct stat_data *sd, i
 	case 'A':
 	{
 		int64_t avg = 0;
-		if (sd->count) {
-			avg = sd->sum;
-			do_div (avg, (int)sd->count); /* FIXME: check for overflow */
-		}
+		if (sd->count)
+			avg = _stp_div64 (NULL, sd->sum, sd->count);
 		_stp_printf("%lld", avg);
 		break;
 	}
@@ -190,7 +214,7 @@ static void _stp_stat_print_valtype (char *fmt, Stat st, struct stat_data *sd, i
 	}
 }
 
-static void __stp_stat_add (Stat st, struct stat_data *sd, int64_t val)
+static void __stp_stat_add (Hist st, stat *sd, int64_t val)
 {
 	int n;
 	if (sd->count == 0) {
@@ -204,22 +228,22 @@ static void __stp_stat_add (Stat st, struct stat_data *sd, int64_t val)
 		if (val < sd->min)
 			sd->min = val;
 	}
-	switch (st->hist_type) {
+	switch (st->type) {
 	case HIST_LOG:
 		n = msb64 (val);
-		if (n >= st->hist_buckets)
-			n = st->hist_buckets - 1;
+		if (n >= st->buckets)
+			n = st->buckets - 1;
 		sd->histogram[n]++;
 		break;
 	case HIST_LINEAR:
-		val -= st->hist_start;
-		do_div (val, st->hist_int);
-		n = val;
-		if (n < 0)
-			n = 0;
-		if (n >= st->hist_buckets)
-			n = st->hist_buckets - 1;
-		sd->histogram[n]++;
+		if (val < st->start)
+			val = st->start;
+		else
+			val -= st->start;
+		do_div (val, st->interval);
+		if (val >= st->buckets)
+			val = st->buckets - 1;
+		sd->histogram[val]++;
 	default:
 		break;
 	}
