@@ -1,4 +1,14 @@
-#ifndef _SYM_C_ /* -*- linux-c -*- */
+/* -*- linux-c -*- 
+ * Symbolic Lookup Functions
+ * Copyright (C) 2005 Red Hat Inc.
+ *
+ * This file is part of systemtap, and is free software.  You can
+ * redistribute it and/or modify it under the terms of the GNU General
+ * Public License (GPL); either version 2, or (at your option) any
+ * later version.
+ */
+
+#ifndef _SYM_C_
 #define _SYM_C_
 
 #include "string.c"
@@ -33,7 +43,6 @@ String _stp_symbol_sprint (String str, unsigned long address)
 		else
 			_stp_sprintf (str, " : %s+%#lx/%#lx", name, offset, size);
 	}
-
 	return str;
 }
 
