@@ -31,8 +31,13 @@ _stp_proc_read (struct file *file, char __user *buf, size_t count, loff_t *ppos)
 		return -EINVAL;
 
 	out.cpu = cpu;
+#ifdef RELAYFS_VERSION_GE_4
+	out.produced = _stp_chan->buf[cpu]->subbufs_produced;
+	out.consumed = _stp_chan->buf[cpu]->subbufs_consumed;
+#else
 	out.produced = atomic_read(&_stp_chan->buf[cpu]->subbufs_produced);
 	out.consumed = atomic_read(&_stp_chan->buf[cpu]->subbufs_consumed);
+#endif /* RELAYFS_VERSION_GE_4 */
 
 	num = sizeof(out);
 	if (copy_to_user(buf, &out, num))
@@ -88,10 +93,7 @@ static ssize_t _stp_proc_write_cmd (struct file *file, const char __user *buf,
 		break;
 	}
 	case STP_EXIT:
-		/* Cannot call _stp_handle_exit() directly here */
-		/* because the buffers may be full and stpd won't be able */
-		/* to empty them until this handler returns. */
-		schedule_work (&stp_exit);
+		_stp_exit_flag = 1;
 		break;
 	case STP_TRANSPORT_INFO:
 	{
@@ -155,15 +157,11 @@ static int _stp_write (int type, void *data, int len)
 	memcpy (bptr->buf, data, len);
 	bptr->len = len;
 	
-
 	/* put it on the pool of ready buffers */
 	spin_lock(&_stp_ready_lock);
 	list_add_tail(&bptr->list, &_stp_ready_q);
 	spin_unlock(&_stp_ready_lock);
 
-	/* now wake up readers */
-	wake_up_interruptible(&_stp_proc_wq);
-
 	return len;
 }
 
@@ -231,7 +229,7 @@ static int _stp_set_buffers(int num)
 	int i;
 	struct list_head *p;
 	
-	printk("stp_set_buffers %d\n", num);
+	//printk("stp_set_buffers %d\n", num);
 
 	if (num == 0 || num == _stp_current_buffers)
 		return _stp_current_buffers;
@@ -239,7 +237,7 @@ static int _stp_set_buffers(int num)
 	spin_lock(&_stp_pool_lock);
 	if (num > _stp_current_buffers) {
 		for (i = 0; i < num - _stp_current_buffers; i++) {
-			p = (struct list_head *)vmalloc(sizeof(struct _stp_buffer));
+			p = (struct list_head *)kmalloc(sizeof(struct _stp_buffer),GFP_KERNEL);
 			if (!p)	{
 				_stp_current_buffers += i;
 				goto err;
@@ -250,7 +248,7 @@ static int _stp_set_buffers(int num)
 		for (i = 0; i < _stp_current_buffers - num; i++) {
 			p = _stp_pool_q.next;
 			list_del(p);
-			vfree(p);
+			kfree(p);
 		}
 	}
 	_stp_current_buffers = num;
@@ -261,9 +259,12 @@ err:
 
 static int _stp_register_procfs (void)
 {
-	int i, j;
+	int i;
 	const char *dirname = "systemtap";
+#ifdef STP_RELAYFS	
+	int j;
 	char buf[8];
+#endif
 	struct proc_dir_entry *de;
 	struct list_head *p, *tmp;
 
@@ -272,7 +273,7 @@ static int _stp_register_procfs (void)
 
 	/* allocate buffers */
 	for (i = 0; i < STP_DEFAULT_BUFFERS; i++) {
-		p = (struct list_head *)vmalloc(sizeof(struct _stp_buffer));
+		p = (struct list_head *)kmalloc(sizeof(struct _stp_buffer),GFP_KERNEL);
 		// printk("allocated buffer at %lx\n", (long)p);
 		if (!p)
 			goto err2;
@@ -327,7 +328,7 @@ static int _stp_register_procfs (void)
 err2:
 	list_for_each_safe(p, tmp, &_stp_pool_q) {
 		list_del(p);
-		vfree(p);
+		kfree(p);
 	}
 
 err1:
@@ -351,11 +352,12 @@ err0:
 
 static void _stp_unregister_procfs (void)
 {
+	struct list_head *p, *tmp;
+#ifdef STP_RELAYFS
 	int i;
 	char buf[8];
-	struct list_head *p, *tmp;
 	struct proc_dir_entry *de;
-#ifdef STP_RELAYFS
+
 	for (de = _stp_proc_mod->subdir; de; de = de->next)
 		kfree (de->data);
 
@@ -370,11 +372,11 @@ static void _stp_unregister_procfs (void)
 	/* free memory pools */
 	list_for_each_safe(p, tmp, &_stp_pool_q) {
 		list_del(p);
-		vfree(p);
+		kfree(p);
 	}
 	list_for_each_safe(p, tmp, &_stp_ready_q) {
 		list_del(p);
-		vfree(p);
+		kfree(p);
 	}
 }
 
